# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: protos/scene.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x12protos/scene.proto\x12\x04mira\"\x1d\n\x05Point\x12\t\n\x01x\x18\x01 \x01(\x05\x12\t\n\x01y\x18\x02 \x01(\x05\"\x9d\x01\n\nAnnotation\x12\x10\n\x08\x63\x61tegory\x18\x01 \x01(\x05\x12\n\n\x02x1\x18\x02 \x01(\x05\x12\n\n\x02x2\x18\x03 \x01(\x05\x12\n\n\x02y1\x18\x04 \x01(\x05\x12\n\n\x02y2\x18\x05 \x01(\x05\x12\x0f\n\x07is_rect\x18\x06 \x01(\x08\x12\x1b\n\x06points\x18\x07 \x03(\x0b\x32\x0b.mira.Point\x12\r\n\x05score\x18\x08 \x01(\x02\x12\x10\n\x08metadata\x18\t \x01(\t\"-\n\x17\x41nnotationConfiguration\x12\x12\n\ncategories\x18\x01 \x03(\t\"C\n\x04Mask\x12\x0f\n\x07visible\x18\x01 \x01(\x08\x12\x0c\n\x04name\x18\x02 \x01(\t\x12\x1c\n\x07\x63ontour\x18\x03 \x03(\x0b\x32\x0b.mira.Point\"\xa4\x01\n\x05Scene\x12\x38\n\x11\x61nnotation_config\x18\x01 \x01(\x0b\x32\x1d.mira.AnnotationConfiguration\x12%\n\x0b\x61nnotations\x18\x02 \x03(\x0b\x32\x10.mira.Annotation\x12\x19\n\x05masks\x18\x03 \x03(\x0b\x32\n.mira.Mask\x12\r\n\x05image\x18\x04 \x01(\x0c\x12\x10\n\x08metadata\x18\x05 \x01(\tb\x06proto3')



_POINT = DESCRIPTOR.message_types_by_name['Point']
_ANNOTATION = DESCRIPTOR.message_types_by_name['Annotation']
_ANNOTATIONCONFIGURATION = DESCRIPTOR.message_types_by_name['AnnotationConfiguration']
_MASK = DESCRIPTOR.message_types_by_name['Mask']
_SCENE = DESCRIPTOR.message_types_by_name['Scene']
Point = _reflection.GeneratedProtocolMessageType('Point', (_message.Message,), {
  'DESCRIPTOR' : _POINT,
  '__module__' : 'protos.scene_pb2'
  # @@protoc_insertion_point(class_scope:mira.Point)
  })
_sym_db.RegisterMessage(Point)

Annotation = _reflection.GeneratedProtocolMessageType('Annotation', (_message.Message,), {
  'DESCRIPTOR' : _ANNOTATION,
  '__module__' : 'protos.scene_pb2'
  # @@protoc_insertion_point(class_scope:mira.Annotation)
  })
_sym_db.RegisterMessage(Annotation)

AnnotationConfiguration = _reflection.GeneratedProtocolMessageType('AnnotationConfiguration', (_message.Message,), {
  'DESCRIPTOR' : _ANNOTATIONCONFIGURATION,
  '__module__' : 'protos.scene_pb2'
  # @@protoc_insertion_point(class_scope:mira.AnnotationConfiguration)
  })
_sym_db.RegisterMessage(AnnotationConfiguration)

Mask = _reflection.GeneratedProtocolMessageType('Mask', (_message.Message,), {
  'DESCRIPTOR' : _MASK,
  '__module__' : 'protos.scene_pb2'
  # @@protoc_insertion_point(class_scope:mira.Mask)
  })
_sym_db.RegisterMessage(Mask)

Scene = _reflection.GeneratedProtocolMessageType('Scene', (_message.Message,), {
  'DESCRIPTOR' : _SCENE,
  '__module__' : 'protos.scene_pb2'
  # @@protoc_insertion_point(class_scope:mira.Scene)
  })
_sym_db.RegisterMessage(Scene)

if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _POINT._serialized_start=28
  _POINT._serialized_end=57
  _ANNOTATION._serialized_start=60
  _ANNOTATION._serialized_end=217
  _ANNOTATIONCONFIGURATION._serialized_start=219
  _ANNOTATIONCONFIGURATION._serialized_end=264
  _MASK._serialized_start=266
  _MASK._serialized_end=333
  _SCENE._serialized_start=336
  _SCENE._serialized_end=500
# @@protoc_insertion_point(module_scope)



import torch
import cv2
import numpy as np
import segmentation_models_pytorch as smp
from .. import detector as mdd
from ... import core as mc


class SMPWrapper(torch.nn.Module):
    """A wrapper for smp models to make them
    compatible with mira detectors."""

    def __init__(self, model, loss):
        super().__init__()
        self.model = model
        self.loss = loss
        self.backbone = self.model.encoder

    def forward(self, x, targets=None):
        """Perform a forward pass, returning loss
        if in training mode, otherwise returning the
        sigmoid outputs."""
        y = self.model(x)
        if self.training:
            return {"loss": self.loss(y_pred=y.contiguous(), y_true=targets)}
        return y.sigmoid()


# pylint: disable=too-many-instance-attributes
class SMP(mdd.Detector):
    """A detector that uses segmentation-models-pytorch
    under the hood to build quasi-object-detection models.

    Args:
        annotation_config: A configuration for annotations.
        arch: A segmentation_models_pytorch model class to use.
            Defaults to smp.UnetPlusPlus.
        device: The device to load the model onto.
        resize_method: The method to use for resizing images.
        encoder_name: The name of the encoder to use with the
            smp model.
        loss: The loss function use. Defaults to
            smp.losses.DiceLoss(smp.losses.MULTILABEL_MODE)
    """

    def __init__(
        self,
        annotation_config: mc.AnnotationConfiguration,
        arch=None,
        device="cpu",
        resize_method="pad_to_multiple",
        encoder_name="efficientnet-b0",
        loss=None,
    ):
        if arch is None:
            arch = smp.UnetPlusPlus
        self.resize_method = resize_method
        self.model = SMPWrapper(
            model=arch(encoder_name=encoder_name, classes=len(annotation_config)),
            loss=loss
            or smp.losses.DiceLoss(
                smp.losses.MULTILABEL_MODE,
            ),
        )
        self.training_model = self.model
        self.backbone = self.model.backbone
        self.set_device(device)
        self.annotation_config = annotation_config
        self.resize_base = 64
        self.preprocessing_fn = smp.encoders.get_preprocessing_fn(
            encoder_name, pretrained="imagenet"
        )
        self.set_input_shape(None, None)

    def compute_inputs(self, images):
        return torch.tensor(
            self.preprocessing_fn(np.float32(images)),
            dtype=torch.float32,
            device=self.device,
        ).permute(0, 3, 1, 2)

    def compute_targets(self, annotation_groups, width, height):
        segmaps = np.zeros(
            (len(annotation_groups), len(self.annotation_config), height, width),
            dtype="float32",
        )
        for annotations, segmap in zip(annotation_groups, segmaps):
            for annotation in annotations:
                index = self.annotation_config.index(annotation.category)
                annotation.draw(segmap[index], color=1, opaque=True)
        return torch.tensor(segmaps, device=self.device)

    @property
    def input_shape(self):
        return self._input_shape

    def invert_targets(self, y, threshold=0.5, **kwargs):
        return [
            mc.utils.flatten(
                [
                    [
                        mc.Annotation(
                            points=contour[:, 0],
                            category=category,
                            score=catmap[
                                contour[:, 0, 1]
                                .min(axis=0) : contour[:, 0, 1]
                                .max(axis=0)
                                + 1,
                                contour[:, 0, 0].min() : contour[:, 0, 0].max() + 1,
                            ].max()
                            if np.product(
                                contour[:, 0].max(axis=0) - contour[:, 0].min(axis=0)
                            )
                            > 0
                            else threshold,
                        )
                        for contour in cv2.findContours(
                            (catmap > threshold).astype("uint8"),
                            mode=cv2.RETR_LIST,
                            method=cv2.CHAIN_APPROX_SIMPLE,
                        )[0]
                    ]
                    for catmap, category in zip(
                        segmap.detach().cpu().numpy(), self.annotation_config
                    )
                ]
            )
            for segmap in y
        ]

    @property
    def serve_module_index(self):
        raise NotImplementedError

    def serve_module_string(self, enable_flexible_size=False):
        raise NotImplementedError

    def set_input_shape(self, width, height):
        self._input_shape = (height, width, 3)



""" EfficientDet Focal, Huber/Smooth L1 loss fns w/ jit support

Based on loss fn in Google's automl EfficientDet repository (Apache 2.0 license).
https://github.com/google/automl/tree/master/efficientdet

Copyright 2020 Ross Wightman
"""
import torch
import torch.nn as nn
import torch.nn.functional as F

from typing import Optional, List, Tuple


def focal_loss_legacy(logits, targets, alpha: float, gamma: float, normalizer):
    """Compute the focal loss between `logits` and the golden `target` values.

    'Legacy focal loss matches the loss used in the official Tensorflow impl for initial
    model releases and some time after that. It eventually transitioned to the 'New' loss
    defined below.

    Focal loss = -(1-pt)^gamma * log(pt)
    where pt is the probability of being classified to the true class.

    Args:
        logits: A float32 tensor of size [batch, height_in, width_in, num_predictions].

        targets: A float32 tensor of size [batch, height_in, width_in, num_predictions].

        alpha: A float32 scalar multiplying alpha to the loss from positive examples
            and (1-alpha) to the loss from negative examples.

        gamma: A float32 scalar modulating loss from hard and easy examples.

         normalizer: A float32 scalar normalizes the total loss from all examples.

    Returns:
        loss: A float32 scalar representing normalized total loss.
    """
    positive_label_mask = targets == 1.0
    cross_entropy = F.binary_cross_entropy_with_logits(logits, targets.to(logits.dtype), reduction='none')
    neg_logits = -1.0 * logits
    modulator = torch.exp(gamma * targets * neg_logits - gamma * torch.log1p(torch.exp(neg_logits)))

    loss = modulator * cross_entropy
    weighted_loss = torch.where(positive_label_mask, alpha * loss, (1.0 - alpha) * loss)
    return weighted_loss / normalizer


def new_focal_loss(logits, targets, alpha: float, gamma: float, normalizer, label_smoothing: float = 0.01):
    """Compute the focal loss between `logits` and the golden `target` values.

    'New' is not the best descriptor, but this focal loss impl matches recent versions of
    the official Tensorflow impl of EfficientDet. It has support for label smoothing, however
    it is a bit slower, doesn't jit optimize well, and uses more memory.

    Focal loss = -(1-pt)^gamma * log(pt)
    where pt is the probability of being classified to the true class.
    Args:
        logits: A float32 tensor of size [batch, height_in, width_in, num_predictions].
        targets: A float32 tensor of size [batch, height_in, width_in, num_predictions].
        alpha: A float32 scalar multiplying alpha to the loss from positive examples
            and (1-alpha) to the loss from negative examples.
        gamma: A float32 scalar modulating loss from hard and easy examples.
        normalizer: Divide loss by this value.
        label_smoothing: Float in [0, 1]. If > `0` then smooth the labels.
    Returns:
        loss: A float32 scalar representing normalized total loss.
    """
    # compute focal loss multipliers before label smoothing, such that it will not blow up the loss.
    pred_prob = logits.sigmoid()
    targets = targets.to(logits.dtype)
    onem_targets = 1. - targets
    p_t = (targets * pred_prob) + (onem_targets * (1. - pred_prob))
    alpha_factor = targets * alpha + onem_targets * (1. - alpha)
    modulating_factor = (1. - p_t) ** gamma

    # apply label smoothing for cross_entropy for each entry.
    if label_smoothing > 0.:
        targets = targets * (1. - label_smoothing) + .5 * label_smoothing
    ce = F.binary_cross_entropy_with_logits(logits, targets, reduction='none')

    # compute the final loss and return
    return (1 / normalizer) * alpha_factor * modulating_factor * ce


def huber_loss(
        input, target, delta: float = 1., weights: Optional[torch.Tensor] = None, size_average: bool = True):
    """
    """
    err = input - target
    abs_err = err.abs()
    quadratic = torch.clamp(abs_err, max=delta)
    linear = abs_err - quadratic
    loss = 0.5 * quadratic.pow(2) + delta * linear
    if weights is not None:
        loss *= weights
    if size_average:
        return loss.mean()
    else:
        return loss.sum()


def smooth_l1_loss(
        input, target, beta: float = 1. / 9, weights: Optional[torch.Tensor] = None, size_average: bool = True):
    """
    very similar to the smooth_l1_loss from pytorch, but with the extra beta parameter
    """
    if beta < 1e-5:
        # if beta == 0, then torch.where will result in nan gradients when
        # the chain rule is applied due to pytorch implementation details
        # (the False branch "0.5 * n ** 2 / 0" has an incoming gradient of
        # zeros, rather than "no gradient"). To avoid this issue, we define
        # small values of beta to be exactly l1 loss.
        loss = torch.abs(input - target)
    else:
        err = torch.abs(input - target)
        loss = torch.where(err < beta, 0.5 * err.pow(2) / beta, err - 0.5 * beta)
    if weights is not None:
        loss *= weights
    if size_average:
        return loss.mean()
    else:
        return loss.sum()


def _box_loss(box_outputs, box_targets, num_positives, delta: float = 0.1):
    """Computes box regression loss."""
    # delta is typically around the mean value of regression target.
    # for instances, the regression targets of 512x512 input with 6 anchors on
    # P3-P7 pyramid is about [0.1, 0.1, 0.2, 0.2].
    normalizer = num_positives * 4.0
    mask = box_targets != 0.0
    box_loss = huber_loss(box_outputs, box_targets, weights=mask, delta=delta, size_average=False)
    return box_loss / normalizer


def one_hot(x, num_classes: int):
    # NOTE: PyTorch one-hot does not handle -ve entries (no hot) like Tensorflow, so mask them out
    x_non_neg = (x >= 0).unsqueeze(-1)
    onehot = torch.zeros(x.shape + (num_classes,), device=x.device, dtype=torch.float32)
    return onehot.scatter(-1, x.unsqueeze(-1) * x_non_neg, 1) * x_non_neg


def loss_fn(
        cls_outputs: List[torch.Tensor],
        box_outputs: List[torch.Tensor],
        cls_targets: List[torch.Tensor],
        box_targets: List[torch.Tensor],
        num_positives: torch.Tensor,
        num_classes: int,
        alpha: float,
        gamma: float,
        delta: float,
        box_loss_weight: float,
        label_smoothing: float = 0.,
        legacy_focal: bool = False,
        ) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
    """Computes total detection loss.
    Computes total detection loss including box and class loss from all levels.
    Args:
        cls_outputs: a List with values representing logits in [batch_size, height, width, num_anchors].
            at each feature level (index)

        box_outputs: a List with values representing box regression targets in
            [batch_size, height, width, num_anchors * 4] at each feature level (index)

        cls_targets: groundtruth class targets.

        box_targets: groundtrusth box targets.

        num_positives: num positive grountruth anchors

    Returns:
        total_loss: an integer tensor representing total loss reducing from class and box losses from all levels.

        cls_loss: an integer tensor representing total class loss.

        box_loss: an integer tensor representing total box regression loss.
    """
    # Sum all positives in a batch for normalization and avoid zero
    # num_positives_sum, which would lead to inf loss during training
    num_positives_sum = (num_positives.sum() + 1.0).float()
    levels = len(cls_outputs)

    cls_losses = []
    box_losses = []
    for l in range(levels):
        cls_targets_at_level = cls_targets[l]
        box_targets_at_level = box_targets[l]

        # Onehot encoding for classification labels.
        cls_targets_at_level_oh = one_hot(cls_targets_at_level, num_classes)

        bs, height, width, _, _ = cls_targets_at_level_oh.shape
        cls_targets_at_level_oh = cls_targets_at_level_oh.view(bs, height, width, -1)
        cls_outputs_at_level = cls_outputs[l].permute(0, 2, 3, 1).float()
        if legacy_focal:
            cls_loss = focal_loss_legacy(
                cls_outputs_at_level, cls_targets_at_level_oh,
                alpha=alpha, gamma=gamma, normalizer=num_positives_sum)
        else:
            cls_loss = new_focal_loss(
                cls_outputs_at_level, cls_targets_at_level_oh,
                alpha=alpha, gamma=gamma, normalizer=num_positives_sum, label_smoothing=label_smoothing)
        cls_loss = cls_loss.view(bs, height, width, -1, num_classes)
        cls_loss = cls_loss * (cls_targets_at_level != -2).unsqueeze(-1)
        cls_losses.append(cls_loss.sum())   # FIXME reference code added a clamp here at some point ...clamp(0, 2))

        box_losses.append(_box_loss(
            box_outputs[l].permute(0, 2, 3, 1).float(),
            box_targets_at_level,
            num_positives_sum,
            delta=delta))

    # Sum per level losses to total loss.
    cls_loss = torch.sum(torch.stack(cls_losses, dim=-1), dim=-1)
    box_loss = torch.sum(torch.stack(box_losses, dim=-1), dim=-1)
    total_loss = cls_loss + box_loss_weight * box_loss
    return total_loss, cls_loss, box_loss


loss_jit = torch.jit.script(loss_fn)


class DetectionLoss(nn.Module):

    __constants__ = ['num_classes']

    def __init__(self, config):
        super(DetectionLoss, self).__init__()
        self.config = config
        self.num_classes = config.num_classes
        self.alpha = config.alpha
        self.gamma = config.gamma
        self.delta = config.delta
        self.box_loss_weight = config.box_loss_weight
        self.label_smoothing = config.label_smoothing
        self.legacy_focal = config.legacy_focal
        self.use_jit = config.jit_loss

    def forward(
            self,
            cls_outputs: List[torch.Tensor],
            box_outputs: List[torch.Tensor],
            cls_targets: List[torch.Tensor],
            box_targets: List[torch.Tensor],
            num_positives: torch.Tensor) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:

        l_fn = loss_fn
        if not torch.jit.is_scripting() and self.use_jit:
            # This branch only active if parent / bench itself isn't being scripted
            # NOTE: I haven't figured out what to do here wrt to tracing, is it an issue?
            l_fn = loss_jit

        return l_fn(
            cls_outputs, box_outputs, cls_targets, box_targets, num_positives,
            num_classes=self.num_classes, alpha=self.alpha, gamma=self.gamma, delta=self.delta,
            box_loss_weight=self.box_loss_weight, label_smoothing=self.label_smoothing, legacy_focal=self.legacy_focal)



# Copyright 2020 Google Research. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Base box coder.

Box coders convert between coordinate frames, namely image-centric
(with (0,0) on the top left of image) and anchor-centric (with (0,0) being
defined by a specific anchor).

Users of a BoxCoder can call two methods:
 encode: which encodes a box with respect to a given anchor
  (or rather, a tensor of boxes wrt a corresponding tensor of anchors) and
 decode: which inverts this encoding with a decode operation.
In both cases, the arguments are assumed to be in 1-1 correspondence already;
it is not the job of a BoxCoder to perform matching.
"""
import torch
from typing import List, Optional
from .box_list import BoxList

# Box coder types.
FASTER_RCNN = 'faster_rcnn'
KEYPOINT = 'keypoint'
MEAN_STDDEV = 'mean_stddev'
SQUARE = 'square'


"""Faster RCNN box coder.

Faster RCNN box coder follows the coding schema described below:
  ty = (y - ya) / ha
  tx = (x - xa) / wa
  th = log(h / ha)
  tw = log(w / wa)
  where x, y, w, h denote the box's center coordinates, width and height
  respectively. Similarly, xa, ya, wa, ha denote the anchor's center
  coordinates, width and height. tx, ty, tw and th denote the anchor-encoded
  center, width and height respectively.

  See http://arxiv.org/abs/1506.01497 for details.
"""


EPS = 1e-8


#@torch.jit.script
class FasterRcnnBoxCoder(object):
    """Faster RCNN box coder."""

    def __init__(self, scale_factors: Optional[List[float]] = None, eps: float = EPS):
        """Constructor for FasterRcnnBoxCoder.

        Args:
            scale_factors: List of 4 positive scalars to scale ty, tx, th and tw.
                If set to None, does not perform scaling. For Faster RCNN,
                the open-source implementation recommends using [10.0, 10.0, 5.0, 5.0].
        """
        self._scale_factors = scale_factors
        if scale_factors is not None:
            assert len(scale_factors) == 4
            for scalar in scale_factors:
                assert scalar > 0
        self.eps = eps

    #@property
    def code_size(self):
        return 4

    def encode(self, boxes: BoxList, anchors: BoxList):
        """Encode a box collection with respect to anchor collection.

        Args:
            boxes: BoxList holding N boxes to be encoded.
            anchors: BoxList of anchors.

        Returns:
            a tensor representing N anchor-encoded boxes of the format [ty, tx, th, tw].
        """
        # Convert anchors to the center coordinate representation.
        ycenter_a, xcenter_a, ha, wa = anchors.get_center_coordinates_and_sizes()
        ycenter, xcenter, h, w = boxes.get_center_coordinates_and_sizes()
        # Avoid NaN in division and log below.
        ha += self.eps
        wa += self.eps
        h += self.eps
        w += self.eps

        tx = (xcenter - xcenter_a) / wa
        ty = (ycenter - ycenter_a) / ha
        tw = torch.log(w / wa)
        th = torch.log(h / ha)
        # Scales location targets as used in paper for joint training.
        if self._scale_factors is not None:
            ty *= self._scale_factors[0]
            tx *= self._scale_factors[1]
            th *= self._scale_factors[2]
            tw *= self._scale_factors[3]
        return torch.stack([ty, tx, th, tw]).t()

    def decode(self, rel_codes, anchors: BoxList):
        """Decode relative codes to boxes.

        Args:
            rel_codes: a tensor representing N anchor-encoded boxes.
            anchors: BoxList of anchors.

        Returns:
            boxes: BoxList holding N bounding boxes.
        """
        ycenter_a, xcenter_a, ha, wa = anchors.get_center_coordinates_and_sizes()

        ty, tx, th, tw = rel_codes.t().unbind()
        if self._scale_factors is not None:
            ty /= self._scale_factors[0]
            tx /= self._scale_factors[1]
            th /= self._scale_factors[2]
            tw /= self._scale_factors[3]
        w = torch.exp(tw) * wa
        h = torch.exp(th) * ha
        ycenter = ty * ha + ycenter_a
        xcenter = tx * wa + xcenter_a
        ymin = ycenter - h / 2.
        xmin = xcenter - w / 2.
        ymax = ycenter + h / 2.
        xmax = xcenter + w / 2.
        return BoxList(torch.stack([ymin, xmin, ymax, xmax]).t())


def batch_decode(encoded_boxes, box_coder: FasterRcnnBoxCoder, anchors: BoxList):
    """Decode a batch of encoded boxes.

    This op takes a batch of encoded bounding boxes and transforms
    them to a batch of bounding boxes specified by their corners in
    the order of [y_min, x_min, y_max, x_max].

    Args:
        encoded_boxes: a float32 tensor of shape [batch_size, num_anchors,
            code_size] representing the location of the objects.
        box_coder: a BoxCoder object.
        anchors: a BoxList of anchors used to encode `encoded_boxes`.

    Returns:
        decoded_boxes: a float32 tensor of shape [batch_size, num_anchors, coder_size]
            representing the corners of the objects in the order of [y_min, x_min, y_max, x_max].

    Raises:
        ValueError: if batch sizes of the inputs are inconsistent, or if
        the number of anchors inferred from encoded_boxes and anchors are inconsistent.
    """
    assert len(encoded_boxes.shape) == 3
    if encoded_boxes.shape[1] != anchors.num_boxes():
        raise ValueError('The number of anchors inferred from encoded_boxes'
                         ' and anchors are inconsistent: shape[1] of encoded_boxes'
                         ' %s should be equal to the number of anchors: %s.' %
                         (encoded_boxes.shape[1], anchors.num_boxes()))

    decoded_boxes = torch.stack([
        box_coder.decode(boxes, anchors).boxes for boxes in encoded_boxes.unbind()
    ])
    return decoded_boxes



# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved
"""
Backbone modules.
"""
from collections import OrderedDict

import torch
import torch.nn.functional as F
import torchvision
from torch import nn
from torchvision.models._utils import IntermediateLayerGetter
from typing import Dict, List

from ..util.misc import NestedTensor, is_main_process

from .position_encoding import build_position_encoding


class FrozenBatchNorm2d(torch.nn.Module):
    """
    BatchNorm2d where the batch statistics and the affine parameters are fixed.

    Copy-paste from torchvision.misc.ops with added eps before rqsrt,
    without which any other models than torchvision.models.resnet[18,34,50,101]
    produce nans.
    """

    def __init__(self, n):
        super(FrozenBatchNorm2d, self).__init__()
        self.register_buffer("weight", torch.ones(n))
        self.register_buffer("bias", torch.zeros(n))
        self.register_buffer("running_mean", torch.zeros(n))
        self.register_buffer("running_var", torch.ones(n))

    def _load_from_state_dict(self, state_dict, prefix, local_metadata, strict,
                              missing_keys, unexpected_keys, error_msgs):
        num_batches_tracked_key = prefix + 'num_batches_tracked'
        if num_batches_tracked_key in state_dict:
            del state_dict[num_batches_tracked_key]

        super(FrozenBatchNorm2d, self)._load_from_state_dict(
            state_dict, prefix, local_metadata, strict,
            missing_keys, unexpected_keys, error_msgs)

    def forward(self, x):
        # move reshapes to the beginning
        # to make it fuser-friendly
        w = self.weight.reshape(1, -1, 1, 1)
        b = self.bias.reshape(1, -1, 1, 1)
        rv = self.running_var.reshape(1, -1, 1, 1)
        rm = self.running_mean.reshape(1, -1, 1, 1)
        eps = 1e-5
        scale = w * (rv + eps).rsqrt()
        bias = b - rm * scale
        return x * scale + bias


class BackboneBase(nn.Module):
    def __init__(self, backbone: nn.Module, return_layer: str):
        super().__init__()
        return_layers = {return_layer: "0"}
        self.body = IntermediateLayerGetter(backbone, return_layers=return_layers)
        self.num_channels = [
            m for m in getattr(backbone, return_layer).modules() if hasattr(m, "out_channels")
        ][-1].out_channels

    def forward(self, tensor_list: NestedTensor):
        xs = self.body(tensor_list.tensors)
        out: Dict[str, NestedTensor] = {}
        for name, x in xs.items():
            m = tensor_list.mask
            assert m is not None
            mask = F.interpolate(m[None].float(), size=x.shape[-2:]).to(torch.bool)[0]
            out[name] = NestedTensor(x, mask)
        return out


class Backbone(BackboneBase):
    """ResNet backbone with frozen BatchNorm."""

    def __init__(self, name: str,
        pretrained_backbone: bool,
        return_layer: str = "layer4",
    ):
        backbone = getattr(torchvision.models, name)(
            pretrained=pretrained_backbone,
            norm_layer=FrozenBatchNorm2d,
        )
        super().__init__(backbone, return_layer)


class Joiner(nn.Sequential):
    def __init__(self, backbone, position_embedding):
        super().__init__(backbone, position_embedding)

    def forward(self, tensor_list: NestedTensor):
        xs = self[0](tensor_list)
        out: List[NestedTensor] = []
        pos = []
        for name, x in xs.items():
            out.append(x)
            # position encoding
            pos.append(self[1](x).to(x.tensors.dtype))

        return out, pos


def build_backbone(args):
    position_embedding = build_position_encoding(args)
    backbone = Backbone(
        name=args.backbone,
        pretrained_backbone=args.pretrained_backbone,
        return_layer=args.return_layer,
    )
    model = Joiner(backbone, position_embedding)
    model.num_channels = backbone.num_channels
    return model



# pylint: disable=too-many-public-methods
import os
import abc
import json
import types
import typing
import random
import tempfile

import torch
import tqdm
import numpy as np
import timm.optim
import timm.scheduler
import pkg_resources
import typing_extensions as tx

try:
    import model_archiver.model_packaging as marmp
    import model_archiver.model_packaging_utils as marmpu
except ImportError:
    marmp = None
    marmpu = None

from .. import metrics as mm
from .. import core as mc
from . import common as mdc

DEFAULT_SCHEDULER_PARAMS = dict(
    sched="cosine",
    min_lr=1e-5,
    decay_rate=0.1,
    warmup_lr=1e-4,
    warmup_epochs=5,
    cooldown_epochs=10,
)

DEFAULT_OPTIMIZER_PARAMS = dict(learning_rate=1e-2, weight_decay=4e-5)

ResizeMethod = tx.Literal["fit", "pad"]


def _loss_from_loss_dict(loss_dict: typing.Dict[str, torch.Tensor]):
    if "loss" in loss_dict:
        return loss_dict["loss"]
    return sum(loss for loss in loss_dict.values())


class Detector:
    """Abstract base class for a detector."""

    @property
    @abc.abstractmethod
    def anchor_boxes(self) -> np.ndarray:
        """Return the list of anchor boxes in xyxy format."""

    model: torch.nn.Module
    backbone: torch.nn.Module
    annotation_config: mc.AnnotationConfiguration
    training_model: torch.nn.Module
    device: typing.Any
    resize_method: tx.Literal["pad", "fit", "pad_to_multiple"]
    resize_base: typing.Optional[int]

    def set_device(self, device):
        """Set the device for training and inference tasks."""
        self.device = torch.device(device)
        self.model.to(self.device)
        if hasattr(self, "training_model"):
            self.training_model.to(self.device)  # type: ignore

    @abc.abstractmethod
    def invert_targets(
        self,
        y: typing.Any,
        threshold: float = 0.5,
        **kwargs,
    ) -> typing.List[typing.List[mc.Annotation]]:
        """Compute a list of annotation groups from model output."""

    def resize_to_model_size(
        self, images: typing.List[np.ndarray]
    ) -> typing.Tuple[np.ndarray, np.ndarray]:
        """Resize a series of images to the current model's size."""
        height, width = self.input_shape[:2]
        padded, scales, _ = mdc.resize(
            images,
            resize_method=self.resize_method,
            height=height,
            width=width,
            base=getattr(self, "resize_base", None),
        )
        # Ensure that scaling maintains aspect ratio.
        np.testing.assert_allclose(*scales.T)
        return padded, scales[:, 0]

    @property
    @abc.abstractmethod
    def input_shape(self) -> typing.Tuple[int, int, int]:
        """Obtain the input shape for this model."""

    @abc.abstractmethod
    def set_input_shape(self, width: int, height: int):
        """Set the input shape for this model."""

    @abc.abstractmethod
    def serve_module_string(self, enable_flexible_size: bool) -> str:
        """Return the module string used as part of TorchServe."""

    @property
    @abc.abstractmethod
    def serve_module_index(self) -> dict:
        """Return the class index -> label mapping for TorchServe."""

    @abc.abstractmethod
    def compute_inputs(self, images: np.ndarray) -> np.ndarray:
        """Convert images into suitable model inputs. *You
        usually should not need this method*. For training,
        use `detector.train()`. For detection, use
        `detector.detect()`.

        Args:
            images: The images to convert

        Returns:
            The input to the model
        """

    @abc.abstractmethod
    def compute_targets(
        self,
        annotation_groups: typing.List[typing.List[mc.Annotation]],
        width: int,
        height: int,
    ) -> typing.Union[typing.List[np.ndarray], np.ndarray]:
        """Compute the expected outputs for a model. *You
        usually should not need this method*. For training,
        use `detector.train()`. For detection, use
        `detector.detect()`.

        Args:
            annotation_groups: A list of lists of annotation groups.

        Returns:
            The output(s) that will be used by detector.train()
        """

    def freeze_backbone(self):
        """Freeze the body of the model, leaving the final classification and
        regression layer as trainable."""
        for p in self.backbone.parameters():  # type: ignore
            p.requires_grad = False
        for m in self.backbone.modules():  # type: ignore
            m.eval()

    def unfreeze_backbone(self, batchnorm=True):
        """Unfreeze the body of the model, making all layers trainable.

        Args:
            batchnorm: Whether to unfreeze batchnorm layers.
        """
        for m in self.backbone.modules():  # type: ignore
            if isinstance(m, torch.nn.BatchNorm2d) and not batchnorm:
                m.eval()
                for p in m.parameters():
                    p.requires_grad = False
            else:
                m.train()
                for p in m.parameters():
                    p.requires_grad = True

    def loss_for_batch(self, batch):
        """Compute the loss for a batch of scenes."""
        assert self.training_model.training, "Model not in training mode."
        images, scales = self.resize_to_model_size(batch.images)
        annotation_groups = [
            [ann.resize(scale) for ann in scene.annotations]
            for scene, scale in zip(batch, scales)
        ]
        return _loss_from_loss_dict(
            self.training_model(
                self.compute_inputs(images),
                self.compute_targets(
                    annotation_groups=annotation_groups,
                    width=images.shape[2],
                    height=images.shape[1],
                ),
            )
        )

    def train(
        self,
        training: mc.SceneCollection,
        validation: mc.SceneCollection = None,
        batch_size: int = 1,
        augmenter: mc.augmentations.AugmenterProtocol = None,
        train_backbone: bool = True,
        train_backbone_bn: bool = True,
        epochs=100,
        shuffle=True,
        callbacks: typing.List[mc.callbacks.CallbackProtocol] = None,
        optimizer_params=None,
        scheduler_params=None,
        clip_grad_norm_params=None,
    ):
        """Run training job.
        Args:
            training: The collection of training images
            validation: The collection of validation images
            batch_size: The batch size to use for training
            augmenter: The augmenter for generating samples
            train_backbone: Whether to fit the backbone.
            epochs: The number of epochs to train.
            shuffle: Whether to shuffle the training data on each epoch.
            callbacks: A list of functions that accept the detector as well
                as a list of previous summaries and returns a dict of summary keys
                and values which will be added to the current summary. It can raise
                StopIteration to stop training early.
            optimizer_params: Passed to timm.optim.create_optimizer_v2 to build
                the optimizer.
            scheduler_params: Passed to timm.schduler.create_scheduler to build
                the scheduler.
        """
        training_model = (
            self.training_model if hasattr(self, "training_model") else self.model
        )
        assert training_model is not None
        optimizer = timm.optim.create_optimizer_v2(
            training_model, **(optimizer_params or DEFAULT_OPTIMIZER_PARAMS)
        )
        scheduler, num_epochs = timm.scheduler.create_scheduler(
            types.SimpleNamespace(
                **{**(scheduler_params or DEFAULT_SCHEDULER_PARAMS), "epochs": epochs}
            ),
            optimizer=optimizer,
        )
        train_index = np.arange(len(training)).tolist()
        summaries = []
        for epoch in range(num_epochs):
            with tqdm.trange(len(training) // batch_size) as t:
                training_model.train()
                if not train_backbone:
                    self.freeze_backbone()
                else:
                    self.unfreeze_backbone(batchnorm=train_backbone_bn)
                t.set_description(f"Epoch {epoch + 1} / {num_epochs}")
                cum_loss = 0
                for batchIdx, start in enumerate(range(0, len(training), batch_size)):
                    if batchIdx == 0 and shuffle:
                        random.shuffle(train_index)
                    end = min(start + batch_size, len(train_index))
                    batch = training.assign(
                        scenes=[training[train_index[idx]] for idx in range(start, end)]
                    )
                    if augmenter is not None:
                        batch = batch.augment(augmenter=augmenter)
                    optimizer.zero_grad()
                    loss = self.loss_for_batch(batch)
                    loss.backward()
                    if clip_grad_norm_params is not None:
                        torch.nn.utils.clip_grad_norm_(
                            training_model.parameters(), **clip_grad_norm_params
                        )
                    cum_loss += loss.detach().cpu().numpy()
                    avg_loss = cum_loss / end
                    optimizer.step()
                    t.set_postfix(loss=avg_loss)
                    t.update()
                summary: typing.Dict[str, typing.Any] = {"loss": avg_loss}
                summaries.append(summary)
                if validation is not None:
                    summary["val_loss"] = np.sum(
                        [
                            self.loss_for_batch(
                                validation.assign(
                                    scenes=[
                                        validation[idx]
                                        for idx in range(
                                            vstart,
                                            min(vstart + batch_size, len(validation)),
                                        )
                                    ]
                                )
                            )
                            .detach()
                            .cpu()
                            .numpy()
                            for vstart in range(0, len(validation), batch_size)
                        ]
                    ) / len(validation)
                scheduler.step(
                    epoch=epoch,
                    metric=avg_loss if validation is None else summary["val_loss"],
                )
                if callbacks is not None:
                    try:
                        for callback in callbacks:
                            for k, v in callback(
                                detector=self, summaries=summaries
                            ).items():
                                summary[k] = v
                    except StopIteration:
                        return summaries
                t.set_postfix(**summary)
        return summaries

    def detect(self, image: np.ndarray, **kwargs) -> typing.List[mc.Annotation]:
        """Run detection for a given image. All other args passed to invert_targets()

        Args:
            image: The image to run detection on

        Returns:
            A list of annotations
        """
        self.model.eval()
        images, scales = self.resize_to_model_size([image])
        with torch.no_grad():
            annotations = self.invert_targets(
                self.model(self.compute_inputs(images)), **kwargs
            )[0]
        return [a.resize(1 / scales[0]) for a in annotations]

    def detect_batch(
        self,
        images: typing.List[np.ndarray],
        batch_size: int = 32,
        **kwargs,
    ) -> typing.List[typing.List[mc.Annotation]]:
        """
        Perform object detection on a batch of images.

        Args:
            images: A list of images
            threshold: The detection threshold for the images
            batch_size: The batch size to use with the underlying model

        Returns:
            A list of lists of annotations.
        """
        self.model.eval()
        annotation_groups = []
        for start in range(0, len(images), batch_size):
            current_images, current_scales = self.resize_to_model_size(
                images[start : start + batch_size]
            )
            annotation_groups.extend(
                [
                    [a.resize(1 / scale) for a in annotations]
                    for scale, annotations in zip(
                        current_scales,
                        self.invert_targets(
                            self.model(
                                self.compute_inputs(current_images),
                            ),
                            **kwargs,
                        ),
                    )
                ]
            )
        return annotation_groups

    def mAP(
        self,
        collection: mc.SceneCollection,
        iou_threshold=0.5,
        min_threshold=0.01,
        batch_size=32,
    ):
        """Compute the mAP metric for a given collection
        of ground truth scenes.

        Args:
            collection: The collection to evaluate
            min_threshold: The minimum threshold for initial selection
                of boxes.
            iou_threshold: The IoU threshold required for
                a match

        Returns:
            mAP score
        """
        pred = collection.assign(
            scenes=[
                scene.assign(annotations=annotations)
                for scene, annotations in zip(
                    collection,
                    self.detect_batch(
                        images=collection.images,
                        threshold=min_threshold,
                        batch_size=batch_size,
                    ),
                )
            ]
        )
        return mm.mAP(
            true_collection=collection,
            pred_collection=pred,
            iou_threshold=iou_threshold,
        )

    def to_torchserve(
        self,
        model_name: str,
        directory=".",
        archive_format: tx.Literal["default", "no-archive"] = "default",
        score_threshold: float = 0.5,
        enable_flexible_size=False,
        model_version="1.0",
    ):
        """Build a TorchServe-compatible MAR file for this model."""
        assert (
            marmpu is not None
        ), "You must `pip install torch-model-archiver` to use this function."
        os.makedirs(directory, exist_ok=True)
        with tempfile.TemporaryDirectory() as tdir:
            serialized_file = os.path.join(tdir, "weights.pth")
            index_to_name_file = os.path.join(tdir, "index_to_name.json")
            model_file = os.path.join(tdir, "model.py")
            handler_file = os.path.join(tdir, "object_detector.py")
            torch.save(self.model.state_dict(prefix="model."), serialized_file)
            with open(index_to_name_file, "w", encoding="utf8") as f:
                f.write(json.dumps(self.serve_module_index))
            with open(model_file, "w", encoding="utf8") as f:
                f.write(
                    self.serve_module_string(enable_flexible_size=enable_flexible_size)
                )
            with open(handler_file, "w", encoding="utf8") as f:
                f.write(
                    pkg_resources.resource_string(
                        "mira", "detectors/assets/serve/object_detector.py"
                    )
                    .decode("utf-8")
                    .replace("SCORE_THRESHOLD", str(score_threshold))  # type: ignore
                )
            args = types.SimpleNamespace(
                model_name=model_name,
                serialized_file=serialized_file,
                handler=handler_file,
                model_file=model_file,
                version=model_version,
                requirements_file=None,
                runtime="python",
                extra_files=index_to_name_file,
                export_path=directory,
                force=True,
                archive_format=archive_format,
            )
            marmp.package_model(
                args=args, manifest=marmpu.ModelExportUtils.generate_manifest_json(args)
            )

    @property
    def anchor_sizes(self):
        """Get an array of anchor sizes (i.e., width and height)."""
        return np.diff(
            self.anchor_boxes.reshape((-1, 2, 2)),
            axis=1,
        )[:, 0, :]



""" PyTorch EfficientDet support benches

Hacked together by Ross Wightman
"""
from typing import Optional, Dict, List
import torch
import torch.nn as nn
from .anchors import Anchors, AnchorLabeler, generate_detections
from .loss import DetectionLoss


def _post_process(
        cls_outputs: List[torch.Tensor],
        box_outputs: List[torch.Tensor],
        num_levels: int,
        num_classes: int,
        max_detection_points: int = 5000,
):
    """Selects top-k predictions.

    Post-proc code adapted from Tensorflow version at: https://github.com/google/automl/tree/master/efficientdet
    and optimized for PyTorch.

    Args:
        cls_outputs: an OrderDict with keys representing levels and values
            representing logits in [batch_size, height, width, num_anchors].

        box_outputs: an OrderDict with keys representing levels and values
            representing box regression targets in [batch_size, height, width, num_anchors * 4].

        num_levels (int): number of feature levels

        num_classes (int): number of output classes
    """
    batch_size = cls_outputs[0].shape[0]
    cls_outputs_all = torch.cat([
        cls_outputs[level].permute(0, 2, 3, 1).reshape([batch_size, -1, num_classes])
        for level in range(num_levels)], 1)

    box_outputs_all = torch.cat([
        box_outputs[level].permute(0, 2, 3, 1).reshape([batch_size, -1, 4])
        for level in range(num_levels)], 1)

    _, cls_topk_indices_all = torch.topk(cls_outputs_all.reshape(batch_size, -1), dim=1, k=max_detection_points)
    indices_all = cls_topk_indices_all // num_classes
    classes_all = cls_topk_indices_all % num_classes

    box_outputs_all_after_topk = torch.gather(
        box_outputs_all, 1, indices_all.unsqueeze(2).expand(-1, -1, 4))

    cls_outputs_all_after_topk = torch.gather(
        cls_outputs_all, 1, indices_all.unsqueeze(2).expand(-1, -1, num_classes))
    cls_outputs_all_after_topk = torch.gather(
        cls_outputs_all_after_topk, 2, classes_all.unsqueeze(2))

    return cls_outputs_all_after_topk, box_outputs_all_after_topk, indices_all, classes_all


@torch.jit.script
def _batch_detection(
        batch_size: int, class_out, box_out, anchor_boxes, indices, classes,
        img_scale: Optional[torch.Tensor] = None,
        img_size: Optional[torch.Tensor] = None,
        max_det_per_image: int = 100,
        soft_nms: bool = False,
):
    batch_detections = []
    # FIXME we may be able to do this as a batch with some tensor reshaping/indexing, PR welcome
    for i in range(batch_size):
        img_scale_i = None if img_scale is None else img_scale[i]
        img_size_i = None if img_size is None else img_size[i]
        detections = generate_detections(
            class_out[i], box_out[i], anchor_boxes, indices[i], classes[i],
            img_scale_i, img_size_i, max_det_per_image=max_det_per_image, soft_nms=soft_nms)
        batch_detections.append(detections)
    return torch.stack(batch_detections, dim=0)


class DetBenchPredict(nn.Module):
    def __init__(self, model):
        super(DetBenchPredict, self).__init__()
        self.model = model
        self.config = model.config  # FIXME remove this when we can use @property (torchscript limitation)
        self.num_levels = model.config.num_levels
        self.num_classes = model.config.num_classes
        self.anchors = Anchors.from_config(model.config)
        self.max_detection_points = model.config.max_detection_points
        self.max_det_per_image = model.config.max_det_per_image
        self.soft_nms = model.config.soft_nms

    def forward(self, x, img_info: Optional[Dict[str, torch.Tensor]] = None):
        class_out, box_out = self.model(x)
        class_out, box_out, indices, classes = _post_process(
            class_out, box_out, num_levels=self.num_levels, num_classes=self.num_classes,
            max_detection_points=self.max_detection_points)
        if img_info is None:
            img_scale, img_size = None, None
        else:
            img_scale, img_size = img_info['img_scale'], img_info['img_size']
        return _batch_detection(
            x.shape[0], class_out, box_out, self.anchors.boxes, indices, classes,
            img_scale, img_size, max_det_per_image=self.max_det_per_image, soft_nms=self.soft_nms
        )


class DetBenchTrain(nn.Module):
    def __init__(self, model, create_labeler=True):
        super(DetBenchTrain, self).__init__()
        self.model = model
        self.config = model.config  # FIXME remove this when we can use @property (torchscript limitation)
        self.num_levels = model.config.num_levels
        self.num_classes = model.config.num_classes
        self.anchors = Anchors.from_config(model.config)
        self.max_detection_points = model.config.max_detection_points
        self.max_det_per_image = model.config.max_det_per_image
        self.soft_nms = model.config.soft_nms
        self.anchor_labeler = None
        if create_labeler:
            self.anchor_labeler = AnchorLabeler(self.anchors, self.num_classes, match_threshold=0.5)
        self.loss_fn = DetectionLoss(model.config)

    def forward(self, x, target: Dict[str, torch.Tensor]):
        class_out, box_out = self.model(x)
        if self.anchor_labeler is None:
            # target should contain pre-computed anchor labels if labeler not present in bench
            assert 'label_num_positives' in target
            cls_targets = [target[f'label_cls_{l}'] for l in range(self.num_levels)]
            box_targets = [target[f'label_bbox_{l}'] for l in range(self.num_levels)]
            num_positives = target['label_num_positives']
        else:
            cls_targets, box_targets, num_positives = self.anchor_labeler.batch_label_anchors(
                target['bbox'], target['cls'])

        loss, class_loss, box_loss = self.loss_fn(class_out, box_out, cls_targets, box_targets, num_positives)
        output = {'loss': loss, 'class_loss': class_loss, 'box_loss': box_loss}
        if not self.training:
            # if eval mode, output detections for evaluation
            class_out_pp, box_out_pp, indices, classes = _post_process(
                class_out, box_out, num_levels=self.num_levels, num_classes=self.num_classes,
                max_detection_points=self.max_detection_points)
            output['detections'] = _batch_detection(
                x.shape[0], class_out_pp, box_out_pp, self.anchors.boxes, indices, classes,
                target['img_scale'], target['img_size'],
                max_det_per_image=self.max_det_per_image, soft_nms=self.soft_nms)
        return output


def unwrap_bench(model):
    # Unwrap a model in support bench so that various other fns can access the weights and attribs of the
    # underlying model directly
    if hasattr(model, 'module'):  # unwrap DDP or EMA
        return unwrap_bench(model.module)
    elif hasattr(model, 'model'):  # unwrap Bench -> model
        return unwrap_bench(model.model)
    else:
        return model



"""Scene and SceneCollection objects"""

# pylint: disable=invalid-name,len-as-condition,unsupported-assignment-operation

import os
import io
import json
import typing
import logging
import tarfile
import tempfile

import tqdm
import pandas as pd
import albumentations as A
import typing_extensions as tx
import matplotlib.pyplot as plt
import matplotlib as mpl
import numpy as np
import cv2

from .protos import scene_pb2 as mps
from .annotation import AnnotationConfiguration, Annotation
from . import utils, augmentations

log = logging.getLogger(__name__)


class Scene:
    """A single annotated image.

    Args:
        annotation_config: The configuration for annotations for the
            image.
        annotations: The list of annotations.
        image: The image that was annotated. Can be lazy-loaded by passing
            a string filepath.
        metadata: Metadata about the scene as a dictionary
        cache: Defines caching behavior for the image. If `True`, image is
            loaded into memory the first time that the image is requested.
            If `False`, image is loaded from the file path or URL whenever
            the image is requested.
        masks: A list of MaskRegion dictonaries which will determine
            which parts of images are shown and hidden.
    """

    def __init__(
        self,
        annotation_config: AnnotationConfiguration,
        annotations: typing.List[Annotation],
        image: typing.Union[np.ndarray, str],
        metadata: dict = None,
        cache: bool = False,
        masks: typing.List[utils.MaskRegion] = None,
    ):
        assert isinstance(
            image, (np.ndarray, str)
        ), "Image must be string or ndarray, not " + str(type(image))
        if masks is None:
            masks = []
        self.metadata = metadata
        self._image = image
        self._annotations = annotations
        self._annotation_config = annotation_config
        self.cache = cache
        self.masks = masks

    @property
    def image(self) -> np.ndarray:
        """The image that is being annotated"""
        # Check to see if we have an actual image
        # or just a string
        protect_image = False
        if isinstance(self._image, str):
            # Load the image
            log.debug("Reading from %s", self._image)
            image = utils.read(self._image)
        else:
            protect_image = True
            log.debug("Reading image from cache.")
            image = self._image
        # Check how to handle caching the image
        # for future reads
        if self.cache is True:
            log.debug("Caching image.")
            protect_image = True
            self._image = image
        elif self.cache is False:
            pass
        else:
            raise ValueError(f"Unsupported cache parameter: {self.cache}.")
        if self.masks:
            if protect_image:
                # We should not modify this image. Work on a copy.
                image = image.copy()
            utils.apply_mask(image, masks=self.masks)
        return image

    @property
    def annotation_config(self):
        """The annotation configuration"""
        return self._annotation_config

    @property
    def annotations(self) -> typing.List[Annotation]:
        """Get the list of annotations"""
        return self._annotations

    @classmethod
    def fromString(cls, string):
        """Deserialize scene from string."""
        deserialized = mps.Scene.FromString(string)
        annotation_config = AnnotationConfiguration(
            deserialized.annotation_config.categories
        )
        image = cv2.imdecode(
            np.frombuffer(deserialized.image, dtype="uint8"), cv2.IMREAD_COLOR
        )
        annotations = []
        for annotation in deserialized.annotations:
            common = {
                "category": annotation_config[annotation.category],
                "metadata": json.loads(annotation.metadata),
            }
            if annotation.is_rect:
                annotations.append(
                    Annotation(
                        x1=annotation.x1,
                        y1=annotation.y1,
                        x2=annotation.x2,
                        y2=annotation.y2,
                        **common,
                    )
                )
            else:
                annotations.append(
                    Annotation(
                        points=np.array([[pt.x, pt.y] for pt in annotation.points]),
                        **common,
                    )
                )
        return cls(
            image=image,
            metadata=json.loads(deserialized.metadata),
            annotations=annotations,
            annotation_config=annotation_config,
            masks=[
                {
                    "visible": m.visible,
                    "name": m.name,
                    "contour": np.array([[p.x, p.y] for p in m.contour]),
                }
                for m in deserialized.masks
            ],
        )

    def toString(self):
        """Serialize scene to string."""
        return mps.Scene(
            image=cv2.imencode(".png", self.image)[1].tobytes(),
            annotation_config=mps.AnnotationConfiguration(
                categories=[c.name for c in self.annotation_config]
            ),
            metadata=json.dumps(self.metadata or {}),
            masks=[
                mps.Mask(
                    visible=m["visible"],
                    name=m["name"],
                    contour=[mps.Point(x=x, y=y) for x, y in m["contour"]],
                )
                for m in (self.masks or [])
            ],
            annotations=[
                mps.Annotation(
                    category=self.annotation_config.index(ann.category),
                    x1=ann.x1,
                    y1=ann.y1,
                    x2=ann.x2,
                    y2=ann.y2,
                    points=[mps.Point(x=x, y=y) for x, y in ann.points],
                    metadata=json.dumps(ann.metadata or {}),
                    is_rect=ann.is_rect,
                )
                for ann in self.annotations
            ],
        ).SerializeToString()

    def assign(self, **kwargs) -> "Scene":
        """Get a new scene with only the supplied
        keyword arguments changed."""
        if "annotation_config" in kwargs:
            # We need to change all the categories for annotations
            # to match the new annotation configuration.
            annotations = kwargs.get("annotations", self.annotations)
            annotation_config = kwargs["annotation_config"]
            revised = [
                ann.convert(annotation_config=annotation_config) for ann in annotations
            ]
            revised = [ann for ann in revised if ann is not None]
            removed = len(annotations) - len(revised)
            log.debug(
                "Removed %s annotations when changing annotation configuration.",
                removed,
            )
            kwargs["annotations"] = revised
        # We use the _image instead of image to avoid triggering an
        # unnecessary read of the actual image.
        defaults = {
            "annotation_config": self.annotation_config,
            "annotations": self.annotations,
            "image": self._image,
            "cache": self.cache,
            "metadata": self.metadata,
            "masks": self.masks,
        }
        kwargs = {**defaults, **kwargs}
        return Scene(**kwargs)

    def show(self, annotation_kwargs=None, **kwargs) -> mpl.axes.Axes:
        """Show an annotated version of the image. All arguments
        passed to `mira.core.utils.imshow()`.
        """
        return utils.imshow(self.annotated(**(annotation_kwargs or {})), **kwargs)

    def scores(self):
        """Obtain an array containing the confidence
        score for each annotation."""
        return np.array([a.score for a in self.annotations])

    def bboxes(self):
        """Obtain an array of shape (N, 5) where the columns are
        x1, y1, x2, y2, class_index where class_index is determined
        from the annotation configuration."""
        # We reshape in order to avoid indexing problems when
        # there are no annotations.
        return self.annotation_config.bboxes_from_group(self.annotations)

    def show_annotations(self, **kwargs):
        """Show annotations as individual plots. All arguments
        passed to plt.subplots."""
        if len(self.annotations) == 0:
            return None
        fig, axs = plt.subplots(nrows=len(self.annotations), **kwargs)
        if len(self.annotations) == 1:
            axs = [axs]
        image = self.image
        for ann, ax in zip(self.annotations, axs):
            ax.imshow(ann.extract(image))
            ax.set_title(ann.category.name)
        return fig

    def drop_duplicates(
        self, threshold=1, method: tx.Literal["iou", "coverage"] = "iou"
    ):
        """Remove annotations of the same class where one annotation covers similar or equal area as another.

        Args:
            method: Whether to check overlap by "coverage" (i.e.,
                is X% of box A contained by some larger box B) or "iou"
                (intersection-over-union). IoU is, of course, more strict.
            threshold: The threshold for equality. Boxes are retained if there
                is no larger box with which the overlap is greater than or
                equal to this threshold.
        """
        annotations = []
        for current_category in self.annotation_config:
            current_annotations = [
                ann for ann in self.annotations if ann.category == current_category
            ]
            # Keep only annotations that are not duplicative with a larger nnotation.
            annotations.extend(
                [
                    current_annotations[idx]
                    for idx in utils.find_largest_unique_boxes(
                        bboxes=self.annotation_config.bboxes_from_group(
                            current_annotations
                        )[:, :4],
                        method=method,
                        threshold=threshold,
                    )
                ]
            )
        return self.assign(annotations=annotations)

    def annotated(
        self, dpi=72, fontsize="x-large", labels=True, opaque=False, color=(255, 0, 0)
    ) -> np.ndarray:
        """Show annotations on the image itself.

        Args:
            dpi: The resolution for the image
            fontsize: How large to show labels
            labels: Whether or not to show labels
            opaque: Whether to draw annotations filled
                in.
            color: The color to use for annotations.
        """
        plt.ioff()
        fig, ax = plt.subplots()
        ax.set_xlabel("")
        ax.set_xticks([])
        ax.set_ylabel("")
        ax.set_yticks([])
        ax.axis("off")
        img_raw = self.image.copy()
        img = img_raw
        for ann in self.annotations:
            ann.draw(img, color=color, opaque=opaque)
        utils.imshow(img, ax=ax)
        if labels:
            for ann in self.annotations:
                x1, y1, _, _ = ann.x1y1x2y2()
                ax.annotate(
                    text=ann.category.name,
                    xy=(x1, y1),
                    fontsize=fontsize,
                    backgroundcolor=(1, 1, 1, 0.5),
                )
        ax.set_xlim(0, img_raw.shape[1])
        ax.set_ylim(img_raw.shape[0], 0)
        fig.canvas.draw()
        raw = io.BytesIO()
        fig.savefig(
            raw,
            dpi=dpi,
            pad_inches=0,
            transparent=False,
            bbox_inches="tight",
        )
        plt.close(fig)
        plt.ion()
        raw.seek(0)
        img = utils.read(raw)
        img = img[:, :, :3]
        raw.close()
        return img

    def augment(
        self, augmenter: augmentations.AugmenterProtocol = None, min_visibility=None
    ) -> "Scene":
        """Obtain an augmented version of the scene using the given augmenter.

        Returns:
            The augmented scene
        """
        if augmenter is None:
            return self
        transformed = augmenter(
            image=self.image,
            bboxes=[ann.x1y1x2y2() for ann in self.annotations],
            bbox_indices=[
                annIdx if ann.is_rect else -1
                for annIdx, ann in enumerate(self.annotations)
            ],
            keypoints=utils.flatten(
                [ann.points.tolist() for ann in self.annotations if not ann.is_rect]
            ),
            keypoint_indices=utils.flatten(
                [
                    [(annIdx, keyIdx) for keyIdx in range(len(ann.points))]
                    for annIdx, ann in enumerate(self.annotations)
                    if not ann.is_rect
                ]
            ),
        )

        image = transformed["image"]
        annotations = [
            Annotation(
                x1=x1,
                y1=y1,
                x2=x2,
                y2=y2,
                category=self.annotations[annIdx].category,
                metadata=self.annotations[annIdx].metadata,
            )
            for (x1, y1, x2, y2), annIdx in zip(
                transformed["bboxes"], transformed["bbox_indices"]
            )
            if annIdx > -1
        ] + [
            Annotation(
                points=keypoints.sort_values("keyIdx")[["x", "y"]].values,
                category=self.annotations[annIdx].category,
                metadata=self.annotations[annIdx].metadata,
            )
            for annIdx, keypoints in pd.concat(
                [
                    pd.DataFrame(
                        transformed["keypoint_indices"], columns=["annIdx", "keyIdx"]
                    ),
                    pd.DataFrame(transformed["keypoints"], columns=["x", "y"]),
                ],
                axis=1,
            ).groupby(["annIdx"])
        ]
        recropped = [
            ann.crop(width=image.shape[1], height=image.shape[0]) for ann in annotations
        ]
        if min_visibility is None:
            min_visibility = (
                augmenter.processors["bboxes"].params.min_visibility
                if isinstance(augmenter, A.Compose)
                else 0.0
            )
        annotations = utils.flatten(
            [
                anns
                for ann, anns in zip(annotations, recropped)
                if ann.area() > 0
                and (sum((a.area() for a in anns), 0) / ann.area()) >= min_visibility
            ]
        )
        annotations = [ann for ann in annotations if ann.area() > 0]
        return self.assign(
            image=image,
            annotations=annotations,
        )

    def to_subcrops(self, max_size: int) -> typing.List["Scene"]:
        """Split a scene into subcrops of some maximum size while trying
        to avoid splitting annotations.

        Args:
            max_size: The maximum size of a crop (it may be smaller at the
                edges of an image).
        """
        if max_size % 2 == 0:
            r1 = r2 = max_size // 2
        else:
            r1 = max_size // 2
            r2 = max_size - r1
        annotations = self.annotations
        if not annotations:
            raise NotImplementedError("This function does not support empty scenes.")
        assert all(
            max(a.x2 - a.x1, a.y2 - a.y1) < max_size for a in annotations
        ), "At least one annotation is too big."
        image = self.image
        ih, iw = image.shape[:2]
        subcrops = []
        captured = []
        for annotation in annotations:
            if annotation in captured:
                # We already captured this annotation.
                continue
            # Get the others, sorted by distance to the
            # current annotation.
            axc, ayc = ((annotation.x1 + annotation.x2) / 2), (
                (annotation.y1 + annotation.y2) / 2
            )
            others = sorted(
                [a for a in annotations if a is not annotation],
                key=lambda a: np.square(
                    [
                        axc - ((a.x1 + a.x2) / 2),  # pylint: disable=cell-var-from-loop
                        ayc - ((a.y1 + a.y2) / 2),  # pylint: disable=cell-var-from-loop
                    ]
                ).sum(),
            )
            solved = False
            for r in range(len(others), -1, -1):
                # Try to fit the annotation and the r-closest other
                # annotations into this crop.
                ann_inc = [annotation] + others[:r]
                ann_exc = [a for a in annotations if a not in ann_inc]
                box_inc = self.annotation_config.bboxes_from_group(ann_inc)[:, :4]
                box_exc = self.annotation_config.bboxes_from_group(ann_exc)[:, :4]
                xmin, ymin = box_inc[:, :2].min(axis=0)
                xmax, ymax = box_inc[:, 2:].max(axis=0)
                if max(xmax - xmin, ymax - ymin) > max_size:
                    # This subset covers too large of an area.
                    continue
                xc, yc = map(
                    lambda v: max(v, r1), [(xmax + xmin) / 2, (ymax + ymin) / 2]
                )
                xc, yc = min(xc, iw - r2), min(yc, ih - r2)
                x1, y1, x2, y2 = map(round, [xc - r1, yc - r1, xc + r2, yc + r2])
                coverages = utils.compute_coverage(
                    np.concatenate([box_inc, box_exc], axis=0),
                    np.array([[x1, y1, x2, y2]]),
                )
                if (coverages[: len(box_inc), 0] == 1).all() and (
                    coverages[len(box_inc) :, 0] == 0
                ).all():
                    captured.extend(ann_inc)
                    subcrops.append(
                        self.assign(
                            image=image[y1:y2, x1:x2],
                            annotations=[
                                a.assign(
                                    x1=a.x1 - x1,
                                    y1=a.y1 - y1,
                                    x2=a.x2 - x1,
                                    y2=a.y2 - y1,
                                )
                                for a in ann_inc
                            ],
                        )
                    )
                    solved = True
                    break
            if not solved:
                raise ValueError("Failed to find a suitable crop.")
        return subcrops


class SceneCollection:
    """A collection of scenes.

    Args:
        annotation_config: The configuration that should be used for all
            underlying scenes.
        scenes: The list of scenes.
    """

    def __init__(
        self,
        scenes: typing.List[Scene],
        annotation_config: AnnotationConfiguration = None,
    ):
        if annotation_config is None:
            annotation_config = scenes[0].annotation_config
        for i, s in enumerate(scenes):
            if s.annotation_config != annotation_config:
                raise ValueError(
                    f"Scene {i+1} of {len(scenes)} has inconsistent configuration."
                )
        self._annotation_config = annotation_config
        self._scenes = scenes

    def __getitem__(self, key):
        return self.scenes[key]

    def __setitem__(self, key, val):
        if key >= len(self.scenes):
            raise ValueError(
                f"Cannot set scene {key} when collection has length {len(self.scenes)}."
            )
        self.scenes[key] = val

    def __len__(self):
        return len(self._scenes)

    def __iter__(self):
        for scene in self._scenes:
            yield scene

    @property
    def scenes(self):
        """The list of scenes"""
        return self._scenes

    @property
    def annotation_config(self):
        """The annotation configuration"""
        return self._annotation_config

    @property
    def annotation_groups(self):
        """The groups of annotations in the collection."""
        return [s.annotations for s in self.scenes]

    @property
    def uniform(self):
        """Specifies whether all scenes in the collection are
        of the same size. Note: This will trigger an image load."""
        return (
            np.unique(np.array([s.image.shape for s in self.scenes]), axis=0).shape[0]
            == 1
        )

    @property
    def annotation_sizes(self):
        """An array of dimensions for the annotations in the collection."""
        return np.diff(
            np.array(
                utils.flatten(
                    [[a.x1y1x2y2() for a in g] for g in self.annotation_groups]
                )
            ).reshape((-1, 2, 2)),
            axis=1,
        )[:, 0, :]

    @property
    def consistent(self):
        """Specifies whether all scenes have the same annotation
        configuration."""
        return all(s.annotation_config == self.annotation_config for s in self.scenes)

    @property
    def images(self):
        """All the images for a scene collection.
        All images will be loaded if not already cached."""
        return [s.image for s in self.scenes]

    def augment(self, augmenter: augmentations.AugmenterProtocol, **kwargs):
        """Obtained an augmented version of the given collection.
        All arguments passed to `Scene.augment`"""
        return self.assign(
            scenes=[s.augment(augmenter=augmenter, **kwargs) for s in self.scenes]
        )

    def split(
        self,
        sizes: typing.List[float],
        random_state: int = 42,
        stratify: typing.Sequence[typing.Hashable] = None,
        group: typing.Sequence[typing.Hashable] = None,
    ) -> typing.Sequence["SceneCollection"]:
        """Obtain new scene collections, split based on a
        given set of proportios.

        For example, to get three collections containing 70%, 15%, and 15% of the
        dataset, respectively, you can do something like the following:

        .. code-block:: python

            training, validation, test = collection.split(
                sizes=[0.7, 0.15, 0.15]
            )

        You can also use the `stratify` argument to ensure an even split
        between different kinds of scenes. For example, to split
        scenes containing at least 3 annotations proportionally,
        do something like the following.

        .. code-block:: python

            training, validation, test = collection.split(
                sizes=[0.7, 0.15, 0.15],
                stratify=[len(s.annotations) >= 3 for s in collection]
            )

        Finally, you can make sure certain scenes end up in the same
        split (e.g., if they're crops from the same base image) using
        the group argument.

        .. code-block:: python

            training, validation, test = collection.split(
                sizes=[0.7, 0.15, 0.15],
                stratify=[len(s.annotations) >= 3 for s in collection],
                group=[s.metadata["origin"] for s in collection]
            )

        Returns:
            A train and test scene collection.
        """
        return [
            self.assign(scenes=scenes)
            for scenes in utils.split(
                self.scenes,
                sizes=sizes,
                random_state=random_state,
                stratify=stratify,
                group=group,
            )
        ]

    def assign(self, **kwargs) -> "SceneCollection":
        """Obtain a new scene with the given keyword arguments
        changing. If `annotation_config` is provided, the annotations
        are converted to the new `annotation_config` first.

        Returns:
            A new scene

        """
        if "annotation_config" in kwargs:
            annotation_config = kwargs["annotation_config"]
            scenes = kwargs.get("scenes", self.scenes)
            kwargs["scenes"] = [
                s.assign(annotation_config=annotation_config) for s in scenes
            ]
        defaults = {"scenes": self.scenes, "annotation_config": self.annotation_config}
        kwargs = {**defaults, **kwargs}
        return SceneCollection(**kwargs)

    def sample(self, n, replace=True) -> "SceneCollection":
        """Get a random subsample of this collection"""
        selected = np.random.choice(len(self.scenes), n, replace=replace)
        return self.assign(scenes=[self.scenes[i] for i in selected])

    def save(self, filename: str):
        """Save scene collection a tarball."""
        with tarfile.open(filename, mode="w") as tar:
            for idx, scene in enumerate(tqdm.tqdm(self.scenes)):
                with tempfile.NamedTemporaryFile() as temp:
                    temp.write(scene.toString())
                    temp.flush()
                    tar.add(name=temp.name, arcname=str(idx))

    @classmethod
    def load(cls, filename: str, directory: str = None):
        """Load scene collection from a tarball. If a directory
        is provided, images will be saved into that directory
        rather than retained in memory."""
        if directory:
            os.makedirs(directory, exist_ok=True)
        scenes = []
        with tarfile.open(filename, mode="r") as tar:
            for idx, member in enumerate(tqdm.tqdm(tar.getmembers())):
                data = tar.extractfile(member)
                if data is None:
                    raise ValueError("Failed to load data from a file in the tarball.")
                scene = Scene.fromString(data.read())
                if directory:
                    image_filepath = os.path.join(directory, f"{idx}.png")
                    cv2.imwrite(image_filepath, scene.image)
                    scene = scene.assign(image=image_filepath)
                scenes.append(scene)
        if len(scenes) == 0:
            raise ValueError("No scenes found.")
        return cls(scenes=scenes, annotation_config=scenes[0].annotation_config)



import types
import typing

import torch
import numpy as np
import typing_extensions as tx


from . import detector
from .. import datasets as mds
from .. import core as mc
from ..thirdparty.detr import models as dm
from ..thirdparty.detr.util import box_ops as dub


class DETRWrapper(torch.nn.Module):
    """A wrapper for DETR training and inference."""

    def __init__(self, model, criterion, postprocessors):
        super().__init__()
        self.model = model
        self.criterion = criterion
        self.postprocessors = postprocessors

    def forward(self, samples, targets=None):
        """Forward pass with return specified by whether we are in training."""
        outputs = self.model(samples)
        if self.training:
            assert targets is not None
            targets_transformed = [
                {
                    # Need cx, cy, w, h, normalized by dimensions.
                    "boxes": dub.box_xyxy_to_cxcywh(t["boxes"])
                    / torch.tensor(
                        samples.shape[-2:][::-1], device=t["boxes"].device
                    ).repeat(2),
                    "labels": t["labels"],
                }
                for t in targets
            ]
            loss_dict = self.criterion(outputs, targets_transformed)
            weight_dict = self.criterion.weight_dict
            return {
                "loss": sum(
                    loss_dict[k] * weight_dict[k]
                    for k in loss_dict.keys()
                    if k in weight_dict
                )
            }
        assert targets is None
        return self.postprocessors["bbox"](
            outputs,
            torch.tensor(
                np.array([s.shape[-2:] for s in samples]), device=samples.device
            ),
        )


DETRBuildArgs = tx.TypedDict(
    "DETRBuildArgs",
    {
        "dataset_file": typing.Optional[str],
        "device": str,
        "num_queries": int,
        "num_classes": int,
        "frozen_weights": typing.Optional[str],
        "masks": bool,
        "bbox_loss_coef": float,
        "giou_loss_coef": float,
        "mask_loss_coef": typing.Optional[float],
        "dice_loss_coef": typing.Optional[float],
        "aux_loss": bool,
        "dec_layers": int,
        "enc_layers": int,
        "eos_coef": float,
        "hidden_dim": int,
        "position_embedding": tx.Literal["sine", "learned"],
        "lr_backbone": float,
        "backbone": str,
        "dilation": bool,
        "pretrained_backbone": bool,
        "dropout": float,
        "nheads": int,
        "dim_feedforward": int,
        "pre_norm": bool,
        "set_cost_class": int,
        "set_cost_bbox": int,
        "set_cost_giou": int,
        "return_layer": str,
    },
)

DEFAULT_DETR_ARGS = DETRBuildArgs(
    dataset_file=None,
    device="cpu",
    num_queries=100,
    frozen_weights=None,
    masks=False,
    bbox_loss_coef=5,
    giou_loss_coef=2,
    mask_loss_coef=None,
    dice_loss_coef=None,
    aux_loss=False,
    dec_layers=6,
    enc_layers=6,
    eos_coef=0.1,
    hidden_dim=256,
    position_embedding="sine",
    lr_backbone=1e-5,
    backbone="resnet50",
    dilation=False,
    pretrained_backbone=False,
    dropout=0.1,
    nheads=8,
    dim_feedforward=2048,
    pre_norm=False,
    set_cost_class=1,
    set_cost_bbox=5,
    set_cost_giou=2,
    num_classes=100,
    return_layer="layer4",
)


class DETR(detector.Detector):
    """Facebook's DETR model."""

    def __init__(
        self,
        annotation_config=mds.COCOAnnotationConfiguration90,
        build_kwargs: DETRBuildArgs = None,
        pretrained_backbone: bool = True,
        pretrained_top=False,
        device="cpu",
        resize_method: detector.ResizeMethod = "fit",
    ):
        if pretrained_top:
            pretrained_backbone = False
        self.build_kwargs = {
            **DEFAULT_DETR_ARGS,
            **(build_kwargs or {}),
            "num_classes": len(annotation_config) + 1,
            "pretrained_backbone": pretrained_backbone,
            "device": device,
        }
        self.resize_method = resize_method
        self.model = DETRWrapper(*dm.build(types.SimpleNamespace(**self.build_kwargs)))
        self.backbone = self.model.model.backbone
        self.annotation_config = annotation_config
        self.set_input_shape(width=512, height=512)
        self.set_device(torch.device(device))
        if pretrained_top:
            assert (
                annotation_config == mds.COCOAnnotationConfiguration90
            ), "Only COCO is supported for pretrained_top=True"
            self.model.model.load_state_dict(
                torch.hub.load_state_dict_from_url(
                    url="https://dl.fbaipublicfiles.com/detr/detr-r50-e632da11.pth",
                    map_location=self.device,
                    check_hash=True,
                )["model"]
            )

    @property
    def training_model(self):
        """The training model."""
        return self.model

    @property
    def anchor_boxes(self):
        return None

    def compute_inputs(self, images):
        mean = np.array([0.485, 0.456, 0.406]) * 255
        std = np.array([0.229, 0.224, 0.225]) * 255
        images = (np.float32(images) - mean) / std
        return (
            torch.tensor(images, dtype=torch.float32)
            .permute(0, 3, 1, 2)
            .to(self.device)
        )

    def compute_targets(self, annotation_groups, width, height):
        return [
            {
                "boxes": torch.tensor(b[:, :4], dtype=torch.float32).to(self.device),
                "labels": torch.tensor(b[:, -1] + 1, dtype=torch.int64).to(self.device),
            }
            for b in [
                self.annotation_config.bboxes_from_group(g) for g in annotation_groups
            ]
        ]

    @property
    def input_shape(self):
        return self._input_shape

    def invert_targets(self, y, threshold=0.5, **kwargs):
        return [
            [
                mc.Annotation(
                    category=self.annotation_config[int(labelIdx) - 1],
                    x1=x1,
                    y1=y1,
                    x2=x2,
                    y2=y2,
                    score=score,
                )
                for (x1, y1, x2, y2), labelIdx, score in zip(
                    labels["boxes"].detach().cpu().numpy(),
                    labels["labels"].detach().cpu().numpy(),
                    labels["scores"].detach().cpu().numpy(),
                )
                if score > threshold
            ]
            for labels in y
        ]

    @property
    def serve_module_index(self):
        raise NotImplementedError

    def serve_module_string(self, enable_flexible_size=False):
        raise NotImplementedError

    def set_input_shape(self, width, height):
        self._input_shape = (height, width, 3)



# Copyright 2020 Google Research. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Base target assigner module.

The job of a TargetAssigner is, for a given set of anchors (bounding boxes) and
groundtruth detections (bounding boxes), to assign classification and regression
targets to each anchor as well as weights to each anchor (specifying, e.g.,
which anchors should not contribute to training loss).

It assigns classification/regression targets by performing the following steps:
1) Computing pairwise similarity between anchors and groundtruth boxes using a
  provided RegionSimilarity Calculator
2) Computing a matching based on the similarity matrix using a provided Matcher
3) Assigning regression targets based on the matching and a provided BoxCoder
4) Assigning classification targets based on the matching and groundtruth labels

Note that TargetAssigners only operate on detections from a single
image at a time, so any logic for applying a TargetAssigner to multiple
images must be handled externally.
"""
import torch
from typing import Optional

from . import box_list
from .region_similarity_calculator import IouSimilarity
from .argmax_matcher import ArgMaxMatcher
from .matcher import Match
from .box_list import BoxList
from .box_coder import FasterRcnnBoxCoder

KEYPOINTS_FIELD_NAME = 'keypoints'


#@torch.jit.script
class TargetAssigner(object):
    """Target assigner to compute classification and regression targets."""

    def __init__(self, similarity_calc: IouSimilarity, matcher: ArgMaxMatcher, box_coder: FasterRcnnBoxCoder,
                 negative_class_weight: float = 1.0, unmatched_cls_target: Optional[float] = None,
                 keypoints_field_name: str = KEYPOINTS_FIELD_NAME):
        """Construct Object Detection Target Assigner.

        Args:
            similarity_calc: a RegionSimilarityCalculator

            matcher: Matcher used to match groundtruth to anchors.

            box_coder: BoxCoder used to encode matching groundtruth boxes with respect to anchors.

            negative_class_weight: classification weight to be associated to negative
                anchors (default: 1.0). The weight must be in [0., 1.].

            unmatched_cls_target: a float32 tensor with shape [d_1, d_2, ..., d_k]
                which is consistent with the classification target for each
                anchor (and can be empty for scalar targets).  This shape must thus be
                compatible with the groundtruth labels that are passed to the "assign"
                function (which have shape [num_gt_boxes, d_1, d_2, ..., d_k]).
                If set to None, unmatched_cls_target is set to be [0] for each anchor.

        Raises:
            ValueError: if similarity_calc is not a RegionSimilarityCalculator or
                if matcher is not a Matcher or if box_coder is not a BoxCoder
        """
        self._similarity_calc = similarity_calc
        self._matcher = matcher
        self._box_coder = box_coder
        self._negative_class_weight = negative_class_weight
        if unmatched_cls_target is not None:
            self._unmatched_cls_target = unmatched_cls_target
        else:
            self._unmatched_cls_target = 0.
        self._keypoints_field_name = keypoints_field_name

    def assign(self, anchors: BoxList, groundtruth_boxes: BoxList, groundtruth_labels=None, groundtruth_weights=None):
        """Assign classification and regression targets to each anchor.

        For a given set of anchors and groundtruth detections, match anchors
        to groundtruth_boxes and assign classification and regression targets to
        each anchor as well as weights based on the resulting match (specifying,
        e.g., which anchors should not contribute to training loss).

        Anchors that are not matched to anything are given a classification target
        of self._unmatched_cls_target which can be specified via the constructor.

        Args:
            anchors: a BoxList representing N anchors

            groundtruth_boxes: a BoxList representing M groundtruth boxes

            groundtruth_labels:  a tensor of shape [M, d_1, ... d_k]
                with labels for each of the ground_truth boxes. The subshape
                [d_1, ... d_k] can be empty (corresponding to scalar inputs).  When set
                to None, groundtruth_labels assumes a binary problem where all
                ground_truth boxes get a positive label (of 1).

            groundtruth_weights: a float tensor of shape [M] indicating the weight to
                assign to all anchors match to a particular groundtruth box. The weights
                must be in [0., 1.]. If None, all weights are set to 1.

            **params: Additional keyword arguments for specific implementations of the Matcher.

        Returns:
            cls_targets: a float32 tensor with shape [num_anchors, d_1, d_2 ... d_k],
                where the subshape [d_1, ..., d_k] is compatible with groundtruth_labels
                which has shape [num_gt_boxes, d_1, d_2, ... d_k].

            cls_weights: a float32 tensor with shape [num_anchors]

            reg_targets: a float32 tensor with shape [num_anchors, box_code_dimension]

            reg_weights: a float32 tensor with shape [num_anchors]

            match: a matcher.Match object encoding the match between anchors and groundtruth boxes,
                with rows corresponding to groundtruth boxes and columns corresponding to anchors.

        Raises:
            ValueError: if anchors or groundtruth_boxes are not of type box_list.BoxList
        """
        if not isinstance(anchors, box_list.BoxList):
            raise ValueError('anchors must be an BoxList')
        if not isinstance(groundtruth_boxes, box_list.BoxList):
            raise ValueError('groundtruth_boxes must be an BoxList')

        # device = anchors.device()
        # if groundtruth_labels is None:
        #     groundtruth_labels = torch.ones(groundtruth_boxes.num_boxes(), device=device).unsqueeze(0)
        #     groundtruth_labels = groundtruth_labels.unsqueeze(-1)
        # if groundtruth_weights is None:
        #     num_gt_boxes = groundtruth_boxes.num_boxes()
        #     if not num_gt_boxes:
        #         num_gt_boxes = groundtruth_boxes.num_boxes()
        #     groundtruth_weights = torch.ones([num_gt_boxes], device=device)

        match_quality_matrix = self._similarity_calc.compare(groundtruth_boxes, anchors)
        match = self._matcher.match(match_quality_matrix)
        reg_targets = self._create_regression_targets(anchors, groundtruth_boxes, match)
        cls_targets = self._create_classification_targets(groundtruth_labels, match)
        #reg_weights = self._create_regression_weights(match, groundtruth_weights)
        #cls_weights = self._create_classification_weights(match, groundtruth_weights)

        return cls_targets, reg_targets, match

    def _create_regression_targets(self, anchors: BoxList, groundtruth_boxes: BoxList, match: Match):
        """Returns a regression target for each anchor.

        Args:
            anchors: a BoxList representing N anchors

            groundtruth_boxes: a BoxList representing M groundtruth_boxes

            match: a matcher.Match object

        Returns:
            reg_targets: a float32 tensor with shape [N, box_code_dimension]
        """
        device = anchors.device()
        zero_box = torch.zeros((1, 4), device=device)
        matched_gt_boxes = match.gather_based_on_match(
            groundtruth_boxes.boxes(), unmatched_value=zero_box, ignored_value=zero_box)
        matched_gt_boxlist = box_list.BoxList(matched_gt_boxes)
        if groundtruth_boxes.has_field(self._keypoints_field_name):
            groundtruth_keypoints = groundtruth_boxes.get_field(self._keypoints_field_name)
            zero_kp = torch.zeros((1,) + groundtruth_keypoints.shape[1:], device=device)
            matched_keypoints = match.gather_based_on_match(
                groundtruth_keypoints, unmatched_value=zero_kp, ignored_value=zero_kp)
            matched_gt_boxlist.add_field(self._keypoints_field_name, matched_keypoints)
        matched_reg_targets = self._box_coder.encode(matched_gt_boxlist, anchors)

        unmatched_ignored_reg_targets = self._default_regression_target(device).repeat(match.match_results.shape[0], 1)

        matched_anchors_mask = match.matched_column_indicator()
        reg_targets = torch.where(matched_anchors_mask.unsqueeze(1), matched_reg_targets, unmatched_ignored_reg_targets)
        return reg_targets

    def _default_regression_target(self, device: torch.device):
        """Returns the default target for anchors to regress to.

        Default regression targets are set to zero (though in this implementation what
        these targets are set to should not matter as the regression weight of any box
        set to regress to the default target is zero).

        Returns:
            default_target: a float32 tensor with shape [1, box_code_dimension]
        """
        return torch.zeros(1, self._box_coder.code_size(), device=device)

    def _create_classification_targets(self, groundtruth_labels, match: Match):
        """Create classification targets for each anchor.

        Assign a classification target of for each anchor to the matching
        groundtruth label that is provided by match.  Anchors that are not matched
        to anything are given the target self._unmatched_cls_target

        Args:
            groundtruth_labels:  a tensor of shape [num_gt_boxes, d_1, ... d_k]
                with labels for each of the ground_truth boxes. The subshape
                [d_1, ... d_k] can be empty (corresponding to scalar labels).
            match: a matcher.Match object that provides a matching between anchors
                and groundtruth boxes.

        Returns:
            a float32 tensor with shape [num_anchors, d_1, d_2 ... d_k], where the
            subshape [d_1, ..., d_k] is compatible with groundtruth_labels which has
            shape [num_gt_boxes, d_1, d_2, ... d_k].
        """
        return match.gather_based_on_match(
            groundtruth_labels,
            unmatched_value=self._unmatched_cls_target, ignored_value=self._unmatched_cls_target)

    def _create_regression_weights(self, match: Match, groundtruth_weights):
        """Set regression weight for each anchor.

        Only positive anchors are set to contribute to the regression loss, so this
        method returns a weight of 1 for every positive anchor and 0 for every
        negative anchor.

        Args:
            match: a matcher.Match object that provides a matching between anchors and groundtruth boxes.
            groundtruth_weights: a float tensor of shape [M] indicating the weight to
                assign to all anchors match to a particular groundtruth box.

        Returns:
            a float32 tensor with shape [num_anchors] representing regression weights.
        """
        return match.gather_based_on_match(groundtruth_weights, ignored_value=0., unmatched_value=0.)

    def _create_classification_weights(self, match: Match, groundtruth_weights):
        """Create classification weights for each anchor.

        Positive (matched) anchors are associated with a weight of
        positive_class_weight and negative (unmatched) anchors are associated with
        a weight of negative_class_weight. When anchors are ignored, weights are set
        to zero. By default, both positive/negative weights are set to 1.0,
        but they can be adjusted to handle class imbalance (which is almost always
        the case in object detection).

        Args:
            match: a matcher.Match object that provides a matching between anchors and groundtruth boxes.
            groundtruth_weights: a float tensor of shape [M] indicating the weight to
                assign to all anchors match to a particular groundtruth box.

        Returns:
            a float32 tensor with shape [num_anchors] representing classification weights.
        """
        return match.gather_based_on_match(
            groundtruth_weights, ignored_value=0., unmatched_value=self._negative_class_weight)

    def box_coder(self):
        """Get BoxCoder of this TargetAssigner.

        Returns:
            BoxCoder object.
        """
        return self._box_coder



# pylint: disable=unsupported-assignment-operation
import io
import os
import typing
import logging
import collections
import urllib
import urllib.request

import typing_extensions
import matplotlib.pyplot as plt
import matplotlib as mpl
import numpy as np
import validators
import cv2

log = logging.getLogger(__name__)

MaskRegion = typing_extensions.TypedDict(
    "MaskRegion", {"visible": bool, "contour": np.ndarray, "name": str}
)


def read(filepath_or_buffer: typing.Union[str, io.BytesIO, typing.BinaryIO]):
    """Read a file into an image object

    Args:
        filepath_or_buffer: The path to the file or any object
            with a `read` method (such as `io.BytesIO`)
    """
    if hasattr(filepath_or_buffer, "read"):
        image = np.asarray(bytearray(filepath_or_buffer.read()), dtype=np.uint8)  # type: ignore
        image = cv2.imdecode(image, cv2.IMREAD_COLOR)
    elif isinstance(filepath_or_buffer, str) and validators.url(filepath_or_buffer):
        with urllib.request.urlopen(filepath_or_buffer) as data:
            return read(data)
    else:
        assert os.path.isfile(
            filepath_or_buffer  # type: ignore
        ), "Could not find image at path: " + str(filepath_or_buffer)
        image = cv2.imread(filepath_or_buffer)
    return cv2.cvtColor(image, cv2.COLOR_BGR2RGB) if len(image.shape) == 3 else image


def get_blank_image(width: int, height: int, n_channels: int, cval=255) -> np.ndarray:
    """Obtain a new blank image with given dimensions.

    Args:
        width: The width of the blank image
        height: The height of the blank image
        n_channels: The number of channels. If 0, the image
            will only have two dimensions (y and x).
        cval: The value to set all pixels to (does not apply to the
            alpha channel)

    Returns:
        The blank image
    """
    if n_channels == 0:
        image = np.zeros((height, width)) + 255
    else:
        image = np.zeros((height, width, n_channels)) + cval
    return image.astype("uint8")


def save(
    image,
    filepath_or_buffer: typing.Union[str, io.BytesIO, typing.BinaryIO],
    extension=".jpg",
):
    """Save the image

    Args:
        filepath_or_buffer: The file or buffer to
            which to save the image. If buffer,
            format must be provided
        esxtension: The extension for the format to use
            if writing to buffer
    """
    image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB) if len(image.shape) == 3 else image
    if hasattr(filepath_or_buffer, "write"):
        data = cv2.imencode(extension, image)[1].tobytes()
        filepath_or_buffer.write(data)  # type: ignore
    else:
        cv2.imwrite(filepath_or_buffer, img=image)


def imshow(image, ax: mpl.axes.Axes = None) -> mpl.axes.Axes:
    """Show an image

    Args:
        ax: Axis on which to show the image

    Returns:
        An axes object
    """
    if ax is None:
        ax = plt
    if len(image.shape) == 3 and image.shape[2] >= 3:
        return ax.imshow(image)
    if len(image.shape) == 3 and image.shape[2] == 1:
        return ax.imshow(image[:, :, 0])
    if len(image.shape) == 2:
        return ax.imshow(image)
    raise ValueError("Incorrect dimensions for image data.")


# pylint: disable=unexpected-keyword-arg
def compute_iou(boxesA: np.ndarray, boxesB: np.ndarray) -> np.ndarray:
    """Compute pairwise IoU for two sets of boxes. Consider the following
    example:
    .. code-block:: python
        compute_iou(
            boxesA=np.array([[0, 0, 1, 1]]),
            boxesB=np.array([[0, 0, 0.5, 0.5], [1, 1, 2, 2]])
        )
    This would yield :code:[[0.25, 0]] because the first box has
    overlap of (0.5 * 0.5) divided by a union area of
    ((0.5 * 0.5) + (1 * 1) - (0.5*0.5)).
    Args:
        boxesA: The first set of boxes, provided as x1, y1, x2, y2
            coordinates as an array with shape (NA, 4).
        boxesB: The second set of boxes provided in same form as boxesA.
    Returns:
        An IoU matrix of shape (NA, NB) where 0 means no overlap and
        1 means complete overlap.
    """

    # A joint matrix for all box pairs so we can vectorize. It has shape
    # (NA, NB, 2, 4).
    boxes = np.zeros((boxesA.shape[0], boxesB.shape[0], 2, 4))
    boxes[:, :, 0] = boxesA[:, np.newaxis, :]
    boxes[:, :, 1] = boxesB[np.newaxis, :, :]
    xA = boxes[..., 0].max(axis=-1)
    yA = boxes[..., 1].max(axis=-1)
    xB = boxes[..., 2].min(axis=-1)
    yB = boxes[..., 3].min(axis=-1)
    interArea = (xB - xA).clip(0) * (yB - yA).clip(0)

    boxAArea = (boxes[..., 0, 2] - boxes[..., 0, 0]) * (
        boxes[..., 0, 3] - boxes[..., 0, 1]
    )
    boxBArea = (boxes[..., 1, 2] - boxes[..., 1, 0]) * (
        boxes[..., 1, 3] - boxes[..., 1, 1]
    )
    iou = interArea / (boxAArea + boxBArea - interArea)
    return iou


def compute_coverage(boxesA: np.ndarray, boxesB: np.ndarray) -> np.ndarray:
    """Compute pairwise overlap of two sets of boxes. Useful for detecting
    redundant bounding boxes where there is no score to consider (as in
    non-max suppression). Consider the following example:

    .. code-block:: python

        compute_overlap(
            boxesA=np.array([[0, 0, 1, 1]]),
            boxesB=np.array([[0, 0, 0.5, 0.5], [1, 1, 2, 2]])
        )
    This would yield :code:[[0.25, 0]] because 0.25 of boxA1 is contained
    within boxB1 while none of boxA1 is contained within box B2. If
    both boxes are for the same class, it may make sense to keep only
    the box that fully encompasses the other (this is problem-specific).
    Note that this function is not symmetric (i.e.,
    compute_coverage(x, y) != compute_coverage(y, x)). In our example,
    if we had reversed the inputs, we would have gotten :code:[[1], [0]]
    because 100% of boxB1 (now A1) is contained in boxA1 (now B1) while
    none of boxB2 (now A2) is contained in boxA1 (now B1).

    Args:
        boxesA: The first set of boxes, provided as x1, y1, x2, y2
            coordinates as an array with shape (NA, 4).
        boxesB: The second set of boxes provided in same form as boxesA.
    Returns:
        An IoU matrix of shape (NA, NB, 2) where 0 means no overlap and
        1 means complete overlap.
    """
    # A joint matrix for all box pairs so we can vectorize. It has shape
    # (NA, NB, 2, 4).
    boxes = np.zeros((boxesA.shape[0], boxesB.shape[0], 2, 4))
    boxes[:, :, 0] = boxesA[:, np.newaxis, :]
    boxes[:, :, 1] = boxesB[np.newaxis, :, :]
    xA = boxes[..., 0].max(axis=-1)
    yA = boxes[..., 1].max(axis=-1)
    xB = boxes[..., 2].min(axis=-1)
    yB = boxes[..., 3].min(axis=-1)
    interArea = (xB - xA).clip(0) * (yB - yA).clip(0)

    boxAArea = (boxes[..., 0, 2] - boxes[..., 0, 0]) * (
        boxes[..., 0, 3] - boxes[..., 0, 1]
    )
    coverageA = interArea / boxAArea
    return coverageA


def split(
    items: typing.List[typing.Any],
    sizes: typing.List[float],
    random_state: int = 42,
    stratify: typing.Sequence[typing.Hashable] = None,
    group: typing.Sequence[typing.Hashable] = None,
) -> typing.Sequence[typing.Any]:
    """Split a list of items into groups of specific proportions.

    For example, to split an array into groups containing 70%, 15%, and 15% of the
    dataset, respectively, you can do something like the following:

    .. code-block:: python

        training, validation, test = split(
            np.arange(100),
            sizes=[0.7, 0.15, 0.15]
        )

    You can also use the `stratify` argument to ensure an even split
    between different kinds of scenes. For example, to split
    scenes containing at least 3 annotations proportionally,
    do something like the following.

    .. code-block:: python

        training, validation, test = split(
            np.arange(100),
            sizes=[0.7, 0.15, 0.15],
            stratify=np.random.choice(np.arange(3), size=100)
        )

    Finally, you can make sure certain scenes end up in the same
    split (e.g., if they're crops from the same base image) using
    the group argument.

    .. code-block:: python

        training, validation, test = split(
            np.arange(100),
            sizes=[0.7, 0.15, 0.15],
            stratify=np.random.choice(np.arange(3), size=100),
            group=np.random.choice(np.arange(20), size=100)
        )

    Returns:
        A train and test scene collection.
    """
    if group is None:
        group = list(range(len(items)))
    if stratify is None:
        stratify = [0] * len(items)
    assert sum(sizes) == 1.0, "The sizes must add up to 1.0."
    assert len(group) == len(items), "group must be the same length as the collection."
    assert len(stratify) == len(
        items
    ), "stratify must be the same length as the collection."
    rng = np.random.default_rng(seed=random_state)
    unique = collections.Counter(group)
    hashes = [
        hash(tuple(set(s for s, g in zip(stratify, group) if g == u))) for u in unique
    ]
    totals = collections.Counter(hashes)
    assert len(unique) >= len(
        sizes
    ), "Cannot group when number of unique groups is less than len(sizes)."
    splits: typing.List[typing.List[typing.Any]] = [[] for _ in range(len(sizes))]
    for ht, t in totals.items():
        for a, u in zip(
            rng.choice(len(sizes), size=t, p=sizes),
            [u for h, u in zip(hashes, unique) if h == ht],
        ):
            splits[a].extend(i for i, g in zip(items, group) if g == u)
    return splits


def flatten(t):
    """Standard utility function for flattening a nested list taken from
    https://stackoverflow.com/questions/952914/how-to-make-a-flat-list-out-of-a-list-of-lists."""
    return [item for sublist in t for item in sublist]


def find_largest_unique_boxes(
    bboxes, threshold=1, method: typing_extensions.Literal["iou", "coverage"] = "iou"
):
    """Find the largest entries in a list of bounding boxes that are not duplicative with
    a larger box.

    Args:
        bboxes: An array of bounding boxes in x1y1x2y2 format.
        method: Whether to check overlap by "coverage" (i.e.,
            is X% of box A contained by some larger box B) or "iou"
            (intersection-over-union). IoU is, of course, more strict.
        threshold: The threshold for equality. Boxes are retained if there
            is no larger box with which the overlap is greater than or
            equal to this threshold.
    """

    assert method in ["iou", "coverage"]
    func = compute_iou if method == "iou" else compute_coverage
    if len(bboxes) <= 1:
        # You can't have duplicates if there's only one or None.
        return np.array([0] if len(bboxes) == 1 else [])

    # Sort by area because we're going to identify duplicates
    # in order of size.
    indexes = np.argsort(np.product(bboxes[:, 2:] - bboxes[:, :2], axis=1))

    # Keep only annotations that are not duplicative with a larger (i.e.,
    # later in our sorted list) annotation. The largest annotation is, of course,
    # always retained.
    return np.array(
        [
            bidx
            for bidx, (cidx, coverages) in zip(
                indexes, enumerate(func(bboxes[indexes], bboxes[indexes]))
            )
            if (cidx == len(bboxes) - 1 or coverages[cidx + 1 :].max() < threshold)
        ]
    )


def apply_mask(image, masks):
    """Given an image and a list of masking contours,
    apply masking in place."""
    if not masks:
        return
    hide = [m for m in masks if not m["visible"]]
    show = [m for m in masks if m["visible"]]
    if show:
        # Assume something is masked, unless it's shown.
        mask = np.ones(image.shape[:2], dtype="uint8")
        for m in show:
            cv2.drawContours(
                mask,
                contours=[m["contour"]],
                contourIdx=-1,
                color=0,
                thickness=-1,
            )
    else:
        # Assume something is unmasked, unless it's hidden.
        mask = np.zeros(image.shape[:2], dtype="uint8")
    for m in hide:
        cv2.drawContours(
            mask,
            contours=[m["contour"]],
            contourIdx=-1,
            color=255,
            thickness=cv2.FILLED,
        )
    image[mask > 0] = 0


def transform_bboxes(
    bboxes: np.ndarray, M: np.ndarray, width: int = None, height: int = None, clip=True
):
    """Transform a set of axis-aligned bounding boxes.

    Args:
        bboxes: An array of shape (N, 4) where each row is (xmin, ymin, xmax, ymax)
        M: A transform matrix of shape 2x3.
        width: The width of the output image (for clipping output boxes)
        height: The height of the output image (for clipping output boxes)
        clip: Whether to apply clipping.
    """
    if len(bboxes) == 0:
        return bboxes
    x1, y1, x2, y2 = bboxes.T
    vertices = np.array([x1, y1, x2, y1, x2, y2, x1, y2]).T.reshape((-1, 4, 2))
    transformed_vertices = cv2.transform(vertices, m=M)
    if clip:
        assert (
            width is not None and height is not None
        ), "If clipping, width and height must be provided."
        transformed_vertices = transformed_vertices.clip(0, [width, height])
    transformed_bboxes = np.concatenate(
        [
            transformed_vertices.min(axis=1).T,
            transformed_vertices.max(axis=1).T,
        ],
        axis=0,
    ).T
    return transformed_bboxes



""" RetinaNet / EfficientDet Anchor Gen

Adapted for PyTorch from Tensorflow impl at
    https://github.com/google/automl/blob/6f6694cec1a48cdb33d5d1551a2d5db8ad227798/efficientdet/anchors.py

Hacked together by Ross Wightman, original copyright below
"""
# Copyright 2020 Google Research. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Anchor definition.

This module is borrowed from TPU RetinaNet implementation:
https://github.com/tensorflow/tpu/blob/master/models/official/retinanet/anchors.py
"""
from typing import Optional, Tuple, Sequence

import numpy as np
import torch
import torch.nn as nn
#import torchvision.ops.boxes as tvb
from torchvision.ops.boxes import batched_nms, remove_small_boxes
from typing import List

from .object_detection import ArgMaxMatcher, FasterRcnnBoxCoder, BoxList, IouSimilarity, TargetAssigner
from .soft_nms import batched_soft_nms


# The minimum score to consider a logit for identifying detections.
MIN_CLASS_SCORE = -5.0

# The score for a dummy detection
_DUMMY_DETECTION_SCORE = -1e5


def decode_box_outputs(rel_codes, anchors, output_xyxy: bool=False):
    """Transforms relative regression coordinates to absolute positions.

    Network predictions are normalized and relative to a given anchor; this
    reverses the transformation and outputs absolute coordinates for the input image.

    Args:
        rel_codes: box regression targets.

        anchors: anchors on all feature levels.

    Returns:
        outputs: bounding boxes.

    """
    ycenter_a = (anchors[:, 0] + anchors[:, 2]) / 2
    xcenter_a = (anchors[:, 1] + anchors[:, 3]) / 2
    ha = anchors[:, 2] - anchors[:, 0]
    wa = anchors[:, 3] - anchors[:, 1]

    ty, tx, th, tw = rel_codes.unbind(dim=1)

    w = torch.exp(tw) * wa
    h = torch.exp(th) * ha
    ycenter = ty * ha + ycenter_a
    xcenter = tx * wa + xcenter_a
    ymin = ycenter - h / 2.
    xmin = xcenter - w / 2.
    ymax = ycenter + h / 2.
    xmax = xcenter + w / 2.
    if output_xyxy:
        out = torch.stack([xmin, ymin, xmax, ymax], dim=1)
    else:
        out = torch.stack([ymin, xmin, ymax, xmax], dim=1)
    return out


def clip_boxes_xyxy(boxes: torch.Tensor, size: torch.Tensor):
    boxes = boxes.clamp(min=0)
    size = torch.cat([size, size], dim=0)
    boxes = boxes.min(size)
    return boxes


def generate_detections(
        cls_outputs, box_outputs, anchor_boxes, indices, classes,
        img_scale: Optional[torch.Tensor], img_size: Optional[torch.Tensor],
        max_det_per_image: int = 100, soft_nms: bool = False):
    """Generates detections with RetinaNet model outputs and anchors.

    Args:
        cls_outputs: a torch tensor with shape [N, 1], which has the highest class
            scores on all feature levels. The N is the number of selected
            top-K total anchors on all levels.

        box_outputs: a torch tensor with shape [N, 4], which stacks box regression
            outputs on all feature levels. The N is the number of selected top-k
            total anchors on all levels.

        anchor_boxes: a torch tensor with shape [N, 4], which stacks anchors on all
            feature levels. The N is the number of selected top-k total anchors on all levels.

        indices: a torch tensor with shape [N], which is the indices from top-k selection.

        classes: a torch tensor with shape [N], which represents the class
            prediction on all selected anchors from top-k selection.

        img_scale: a float tensor representing the scale between original image
            and input image for the detector. It is used to rescale detections for
            evaluating with the original groundtruth annotations.

        max_det_per_image: an int constant, added as argument to make torchscript happy

    Returns:
        detections: detection results in a tensor with shape [max_det_per_image, 6],
            each row representing [x_min, y_min, x_max, y_max, score, class]
    """
    assert box_outputs.shape[-1] == 4
    assert anchor_boxes.shape[-1] == 4
    assert cls_outputs.shape[-1] == 1

    anchor_boxes = anchor_boxes[indices, :]

    # Appply bounding box regression to anchors, boxes are converted to xyxy
    # here since PyTorch NMS expects them in that form.
    boxes = decode_box_outputs(box_outputs.float(), anchor_boxes, output_xyxy=True)
    if img_scale is not None and img_size is not None:
        boxes = clip_boxes_xyxy(boxes, img_size / img_scale)  # clip before NMS better?

    scores = cls_outputs.sigmoid().squeeze(1).float()
    if soft_nms:
        top_detection_idx, soft_scores = batched_soft_nms(
            boxes, scores, classes, method_gaussian=True, iou_threshold=0.3, score_threshold=.001)
        scores[top_detection_idx] = soft_scores
    else:
        top_detection_idx = batched_nms(boxes, scores, classes, iou_threshold=0.5)

    # keep only top max_det_per_image scoring predictions
    top_detection_idx = top_detection_idx[:max_det_per_image]
    boxes = boxes[top_detection_idx]
    scores = scores[top_detection_idx, None]
    classes = classes[top_detection_idx, None] + 1  # back to class idx with background class = 0

    if img_scale is not None:
        boxes = boxes * img_scale

    # FIXME add option to convert boxes back to yxyx? Otherwise must be handled downstream if
    # that is the preferred output format.

    # stack em and pad out to max_det_per_image if necessary
    num_det = len(top_detection_idx)
    detections = torch.cat([boxes, scores, classes.float()], dim=1)
    if num_det < max_det_per_image:
        detections = torch.cat([
            detections,
            torch.zeros((max_det_per_image - num_det, 6), device=detections.device, dtype=detections.dtype)
        ], dim=0)
    return detections


def get_feat_sizes(image_size: Tuple[int, int], max_level: int):
    """Get feat widths and heights for all levels.
    Args:
      image_size: a tuple (H, W)
      max_level: maximum feature level.
    Returns:
      feat_sizes: a list of tuples (height, width) for each level.
    """
    feat_size = image_size
    feat_sizes = [feat_size]
    for _ in range(1, max_level + 1):
        feat_size = ((feat_size[0] - 1) // 2 + 1, (feat_size[1] - 1) // 2 + 1)
        feat_sizes.append(feat_size)
    return feat_sizes


class Anchors(nn.Module):
    """RetinaNet Anchors class."""

    def __init__(self, min_level, max_level, num_scales, aspect_ratios, anchor_scale, image_size: Tuple[int, int]):
        """Constructs multiscale RetinaNet anchors.

        Args:
            min_level: integer number of minimum level of the output feature pyramid.

            max_level: integer number of maximum level of the output feature pyramid.

            num_scales: integer number representing intermediate scales added
                on each level. For instances, num_scales=2 adds two additional
                anchor scales [2^0, 2^0.5] on each level.

            aspect_ratios: list of tuples representing the aspect ratio anchors added
                on each level. For instances, aspect_ratios =
                [(1, 1), (1.4, 0.7), (0.7, 1.4)] adds three anchors on each level.

            anchor_scale: float number representing the scale of size of the base
                anchor to the feature stride 2^level.

            image_size: Sequence specifying input image size of model (H, W).
                The image_size should be divided by the largest feature stride 2^max_level.
        """
        super(Anchors, self).__init__()
        self.min_level = min_level
        self.max_level = max_level
        self.num_scales = num_scales
        self.aspect_ratios = aspect_ratios
        if isinstance(anchor_scale, Sequence):
            assert len(anchor_scale) == max_level - min_level + 1
            self.anchor_scales = anchor_scale
        else:
            self.anchor_scales = [anchor_scale] * (max_level - min_level + 1)

        assert isinstance(image_size, Sequence) and len(image_size) == 2
        # FIXME this restriction can likely be relaxed with some additional changes
        assert image_size[0] % 2 ** max_level == 0, 'Image size must be divisible by 2 ** max_level (128)'
        assert image_size[1] % 2 ** max_level == 0, 'Image size must be divisible by 2 ** max_level (128)'
        self.image_size = tuple(image_size)
        self.feat_sizes = get_feat_sizes(image_size, max_level)
        self.config = self._generate_configs()
        self.register_buffer('boxes', self._generate_boxes())

    @classmethod
    def from_config(cls, config):
        return cls(
            config.min_level, config.max_level,
            config.num_scales, config.aspect_ratios,
            config.anchor_scale, config.image_size)

    def _generate_configs(self):
        """Generate configurations of anchor boxes."""
        anchor_configs = {}
        feat_sizes = self.feat_sizes
        for level in range(self.min_level, self.max_level + 1):
            anchor_configs[level] = []
            for scale_octave in range(self.num_scales):
                for aspect in self.aspect_ratios:
                    anchor_configs[level].append(
                        ((feat_sizes[0][0] // feat_sizes[level][0],
                          feat_sizes[0][1] // feat_sizes[level][1]),
                         scale_octave / float(self.num_scales), aspect,
                         self.anchor_scales[level - self.min_level]))
        return anchor_configs

    def _generate_boxes(self):
        """Generates multiscale anchor boxes."""
        boxes_all = []
        for _, configs in self.config.items():
            boxes_level = []
            for config in configs:
                stride, octave_scale, aspect, anchor_scale = config
                base_anchor_size_x = anchor_scale * stride[1] * 2 ** octave_scale
                base_anchor_size_y = anchor_scale * stride[0] * 2 ** octave_scale
                if isinstance(aspect, Sequence):
                    aspect_x = aspect[0]
                    aspect_y = aspect[1]
                else:
                    aspect_x = np.sqrt(aspect)
                    aspect_y = 1.0 / aspect_x
                anchor_size_x_2 = base_anchor_size_x * aspect_x / 2.0
                anchor_size_y_2 = base_anchor_size_y * aspect_y / 2.0

                x = np.arange(stride[1] / 2, self.image_size[1], stride[1])
                y = np.arange(stride[0] / 2, self.image_size[0], stride[0])
                xv, yv = np.meshgrid(x, y)
                xv = xv.reshape(-1)
                yv = yv.reshape(-1)

                boxes = np.vstack((yv - anchor_size_y_2, xv - anchor_size_x_2,
                                   yv + anchor_size_y_2, xv + anchor_size_x_2))
                boxes = np.swapaxes(boxes, 0, 1)
                boxes_level.append(np.expand_dims(boxes, axis=1))

            # concat anchors on the same level to the reshape NxAx4
            boxes_level = np.concatenate(boxes_level, axis=1)
            boxes_all.append(boxes_level.reshape([-1, 4]))

        anchor_boxes = np.vstack(boxes_all)
        anchor_boxes = torch.from_numpy(anchor_boxes).float()
        return anchor_boxes

    def get_anchors_per_location(self):
        return self.num_scales * len(self.aspect_ratios)


class AnchorLabeler(object):
    """Labeler for multiscale anchor boxes.
    """

    def __init__(self, anchors, num_classes: int, match_threshold: float = 0.5):
        """Constructs anchor labeler to assign labels to anchors.

        Args:
            anchors: an instance of class Anchors.

            num_classes: integer number representing number of classes in the dataset.

            match_threshold: float number between 0 and 1 representing the threshold
                to assign positive labels for anchors.
        """
        similarity_calc = IouSimilarity()
        matcher = ArgMaxMatcher(
            match_threshold,
            unmatched_threshold=match_threshold,
            negatives_lower_than_unmatched=True,
            force_match_for_each_row=True)
        box_coder = FasterRcnnBoxCoder()

        self.target_assigner = TargetAssigner(similarity_calc, matcher, box_coder)
        self.anchors = anchors
        self.match_threshold = match_threshold
        self.num_classes = num_classes
        self.indices_cache = {}

    def label_anchors(self, gt_boxes, gt_classes, filter_valid=True):
        """Labels anchors with ground truth inputs.

        Args:
            gt_boxes: A float tensor with shape [N, 4] representing groundtruth boxes.
                For each row, it stores [y0, x0, y1, x1] for four corners of a box.

            gt_classes: A integer tensor with shape [N, 1] representing groundtruth classes.

            filter_valid: Filter out any boxes w/ gt class <= -1 before assigning

        Returns:
            cls_targets_dict: ordered dictionary with keys [min_level, min_level+1, ..., max_level].
                The values are tensor with shape [height_l, width_l, num_anchors]. The height_l and width_l
                represent the dimension of class logits at l-th level.

            box_targets_dict: ordered dictionary with keys [min_level, min_level+1, ..., max_level].
                The values are tensor with shape [height_l, width_l, num_anchors * 4]. The height_l and
                width_l represent the dimension of bounding box regression output at l-th level.

            num_positives: scalar tensor storing number of positives in an image.
        """
        cls_targets_out = []
        box_targets_out = []

        if filter_valid:
            valid_idx = gt_classes > -1  # filter gt targets w/ label <= -1
            gt_boxes = gt_boxes[valid_idx]
            gt_classes = gt_classes[valid_idx]

        cls_targets, box_targets, matches = self.target_assigner.assign(
            BoxList(self.anchors.boxes), BoxList(gt_boxes), gt_classes)

        # class labels start from 1 and the background class = -1
        cls_targets = (cls_targets - 1).long()

        # Unpack labels.
        """Unpacks an array of cls/box into multiple scales."""
        count = 0
        for level in range(self.anchors.min_level, self.anchors.max_level + 1):
            feat_size = self.anchors.feat_sizes[level]
            steps = feat_size[0] * feat_size[1] * self.anchors.get_anchors_per_location()
            cls_targets_out.append(cls_targets[count:count + steps].view([feat_size[0], feat_size[1], -1]))
            box_targets_out.append(box_targets[count:count + steps].view([feat_size[0], feat_size[1], -1]))
            count += steps

        num_positives = (matches.match_results > -1).float().sum()

        return cls_targets_out, box_targets_out, num_positives

    def batch_label_anchors(self, gt_boxes, gt_classes, filter_valid=True):
        batch_size = len(gt_boxes)
        assert batch_size == len(gt_classes)
        num_levels = self.anchors.max_level - self.anchors.min_level + 1
        cls_targets_out = [[] for _ in range(num_levels)]
        box_targets_out = [[] for _ in range(num_levels)]
        num_positives_out = []

        anchor_box_list = BoxList(self.anchors.boxes)
        for i in range(batch_size):
            last_sample = i == batch_size - 1

            if filter_valid:
                valid_idx = gt_classes[i] > -1  # filter gt targets w/ label <= -1
                gt_box_list = BoxList(gt_boxes[i][valid_idx])
                gt_class_i = gt_classes[i][valid_idx]
            else:
                gt_box_list = BoxList(gt_boxes[i])
                gt_class_i = gt_classes[i]
            cls_targets, box_targets, matches = self.target_assigner.assign(anchor_box_list, gt_box_list, gt_class_i)

            # class labels start from 1 and the background class = -1
            cls_targets = (cls_targets - 1).long()

            # Unpack labels.
            """Unpacks an array of cls/box into multiple scales."""
            count = 0
            for level in range(self.anchors.min_level, self.anchors.max_level + 1):
                level_idx = level - self.anchors.min_level
                feat_size = self.anchors.feat_sizes[level]
                steps = feat_size[0] * feat_size[1] * self.anchors.get_anchors_per_location()
                cls_targets_out[level_idx].append(
                    cls_targets[count:count + steps].view([feat_size[0], feat_size[1], -1]))
                box_targets_out[level_idx].append(
                    box_targets[count:count + steps].view([feat_size[0], feat_size[1], -1]))
                count += steps
                if last_sample:
                    cls_targets_out[level_idx] = torch.stack(cls_targets_out[level_idx])
                    box_targets_out[level_idx] = torch.stack(box_targets_out[level_idx])

            num_positives_out.append((matches.match_results > -1).float().sum())
            if last_sample:
                num_positives_out = torch.stack(num_positives_out)

        return cls_targets_out, box_targets_out, num_positives_out




"""COCO parsing tools"""
# pylint: disable=invalid-name

import logging
from os import path
import json

from tqdm import tqdm
import numpy as np

from ..core import (
    Scene,
    SceneCollection,
    AnnotationConfiguration,
    Annotation,
)

log = logging.getLogger(__name__)


def load_coco(
    annotations_file: str,
    image_dir: str,
    annotation_config: AnnotationConfiguration = None,
) -> SceneCollection:
    """Obtain a scene collection from a COCO JSON file.

    Args:
        annotations_file: The annotation file to load
        image_dir: The directory in which to look for images
        annotation_config: The annotation configuration to
            use. If None, it is inferred from the annotations
            category file.

    Returns:
        A scene collection
    """
    if image_dir is None:
        image_dir = path.split(annotations_file)[0]
    with open(annotations_file, "r", encoding="utf8") as f:
        data = json.load(f)
    categories = {}
    for category in data["categories"]:
        categories[category["id"]] = category["name"]
    category_names = [
        c[1] for c in sorted(list(categories.items()), key=lambda x: x[0])
    ]
    if annotation_config is None:
        annotation_config = AnnotationConfiguration(category_names)
    assert len(categories) == len(
        annotation_config
    ), "Annotation configuration incompatible with in-file categories"
    assert all(
        c in annotation_config for c in category_names
    ), "Some in-file categories not in annotation configuration"
    assert all(
        c.name in category_names for c in annotation_config
    ), "Some annotation configuration categories not in file"

    annotations = np.array(
        [
            [ann["image_id"], ann["category_id"]] + ann["bbox"]
            for ann in data["annotations"]
        ]
    )
    annotations = annotations[annotations[:, 0].argsort()]
    images = sorted(data["images"], key=lambda x: x["id"])
    del data
    scenes = []
    startIdx = 0
    for image in tqdm(images, total=len(images), desc="Creating scenes"):
        current = annotations[startIdx:][annotations[startIdx:, 0] == image["id"], 1:]
        startIdx += len(current)
        scenes.append(
            Scene(
                image=path.join(image_dir, image["file_name"]),
                annotation_config=annotation_config,
                annotations=[
                    Annotation(
                        category=annotation_config[categories[int(ann[0])]],
                        x1=ann[1],
                        y1=ann[2],
                        x2=ann[1] + ann[3],
                        y2=ann[2] + ann[4],
                    )
                    for ann in current
                ],
            )
        )
    return SceneCollection(scenes=scenes, annotation_config=annotation_config)


def load_coco_text(
    annotations_file: str,
    image_dir: str,
    annotation_config: AnnotationConfiguration = None,
) -> SceneCollection:
    """Obtain a scene collection from a COCO Text JSON file
    (e.g., that which can be obtained from https://bgshih.github.io/cocotext/)

    Args:
        annotations_file: The annotation file to load
        image_dir: The directory in which to look for images
        annotation_config: The annotation configuration to
            use. If None, it is inferred from the annotations
            category file.

    Returns:
        A scene collection
    """
    with open(annotations_file, "r", encoding="utf8") as f:
        data = json.load(f)

    category_names = set(ann["class"] for ann in data["anns"].values())
    if annotation_config is None:
        annotation_config = AnnotationConfiguration(sorted(list(category_names)))
    assert len(category_names) == len(
        annotation_config
    ), "Annotation configuration incompatible with in-file categories"
    assert all(
        c in annotation_config for c in category_names
    ), "Some in-file categories not in annotation configuration"
    assert all(
        c.name in category_names for c in annotation_config
    ), "Some annotation configuration categories not in file"

    images = data["imgs"]
    scenes = []
    for (imageId, imageData) in tqdm(
        images.items(), total=len(images), desc="Creating scenes"
    ):
        anns = [data["anns"][str(annId)] for annId in data["imgToAnns"][imageId]]
        scenes.append(
            Scene(
                annotation_config=annotation_config,
                image=path.join(image_dir, imageData["file_name"]),
                annotations=[
                    Annotation(
                        category=annotation_config[ann["class"]],
                        x1=ann["bbox"][0],
                        y1=ann["bbox"][1],
                        x2=ann["bbox"][0] + ann["bbox"][2],
                        y2=ann["bbox"][1] + ann["bbox"][3],
                    )
                    for ann in anns
                ],
            )
        )
    return SceneCollection(scenes=scenes, annotation_config=annotation_config)



# type: ignore
# pylint: disable=exec-used,missing-class-docstring,missing-function-docstring,too-few-public-methods
import types
import typing
import zipfile
import json

import ts
import torch
from ts.torch_handler.unit_tests.test_utils.mock_context import MockContext


def import_code(code, name):
    """Import code from a string into a module."""
    module = types.ModuleType(name)
    exec(code, module.__dict__)
    return module


def load_mar_model(filepath: str, gpu_id: int = None):
    """Load a PyTorch model into a customized handler. Used for testing
    handlers to make sure they produce the same output locally as when serving.

    Args:
        filepath: Path to the .MAR file.
        gpu_id: When using a GPU, the ID of the GPU to use (e.g., 0).
    """
    with zipfile.ZipFile(filepath) as z:
        with z.open("MAR-INF/MANIFEST.json", "r") as f:
            manifest = json.loads(f.read().decode("utf-8"))
        handler_name = manifest["model"]["handler"]
        if handler_name.endswith(".py"):
            with z.open(manifest["model"]["handler"]) as f:
                handler_module = import_code(
                    f.read(), manifest["model"]["handler"].split(".py")[0]
                )
        else:
            handler_module = getattr(ts.torch_handler, handler_name)
        candidate_handlers = [
            cls
            for name, cls in handler_module.__dict__.items()
            if isinstance(cls, type)
            and issubclass(cls, ts.torch_handler.vision_handler.VisionHandler)
        ]
        with z.open(manifest["model"]["modelFile"]) as f:
            model_module = import_code(
                f.read(), manifest["model"]["modelFile"].split(".py")[0]
            )
        candidate_models = [
            cls
            for name, cls in model_module.__dict__.items()
            if isinstance(cls, type) and issubclass(cls, torch.nn.Module)
        ]
        if len(candidate_handlers) == 0:
            raise ValueError("No handler found.")
        if len(candidate_models) == 0:
            raise ValueError("No model found.")

        class SimulatedHandler(candidate_handlers[-1]):
            def handle_files(self, filepaths: typing.List[str]):
                data = []
                for filepath in filepaths:
                    with open(filepath, "rb") as f:
                        data.append({"data": f.read()})
                return self.handle(data=data, context=MockContext())

        handler = SimulatedHandler()
        handler.model = candidate_models[-1]()
        # This replaces the "initialize" step in the handler.
        handler.map_location = (
            "cuda" if torch.cuda.is_available() and gpu_id is not None else "cpu"
        )
        handler.device = torch.device(
            handler.map_location + ":" + str(gpu_id)
            if torch.cuda.is_available() and gpu_id is not None
            else handler.map_location
        )
        handler.model.to(handler.device)
        with z.open(manifest["model"]["serializedFile"]) as f:
            handler.model.load_state_dict(
                torch.load(f, map_location=handler.map_location)
            )
        with z.open("index_to_name.json") as f:
            handler.mapping = json.loads(f.read().decode("utf8"))
        handler.model.eval()
        handler.initialized = True
        return handler



# Copyright 2020 Google Research. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Argmax matcher implementation.

This class takes a similarity matrix and matches columns to rows based on the
maximum value per column. One can specify matched_thresholds and
to prevent columns from matching to rows (generally resulting in a negative
training example) and unmatched_theshold to ignore the match (generally
resulting in neither a positive or negative training example).

This matcher is used in Fast(er)-RCNN.

Note: matchers are used in TargetAssigners. There is a create_target_assigner
factory function for popular implementations.
"""
import torch
from .matcher import Match
from typing import Optional


def one_hot_bool(x, num_classes: int):
    # for improved perf over PyTorch builtin one_hot, scatter to bool
    onehot = torch.zeros(x.size(0), num_classes, device=x.device, dtype=torch.bool)
    return onehot.scatter_(1, x.unsqueeze(1), 1)


@torch.jit.script
class ArgMaxMatcher(object):  # cannot inherit with torchscript
    """Matcher based on highest value.

    This class computes matches from a similarity matrix. Each column is matched
    to a single row.

    To support object detection target assignment this class enables setting both
    matched_threshold (upper threshold) and unmatched_threshold (lower thresholds)
    defining three categories of similarity which define whether examples are
    positive, negative, or ignored:
    (1) similarity >= matched_threshold: Highest similarity. Matched/Positive!
    (2) matched_threshold > similarity >= unmatched_threshold: Medium similarity.
            Depending on negatives_lower_than_unmatched, this is either
            Unmatched/Negative OR Ignore.
    (3) unmatched_threshold > similarity: Lowest similarity. Depending on flag
            negatives_lower_than_unmatched, either Unmatched/Negative OR Ignore.
    For ignored matches this class sets the values in the Match object to -2.
    """

    def __init__(self,
                 matched_threshold: float,
                 unmatched_threshold: Optional[float] = None,
                 negatives_lower_than_unmatched: bool = True,
                 force_match_for_each_row: bool = False):
        """Construct ArgMaxMatcher.

        Args:
            matched_threshold: Threshold for positive matches. Positive if
                sim >= matched_threshold, where sim is the maximum value of the
                similarity matrix for a given column. Set to None for no threshold.
            unmatched_threshold: Threshold for negative matches. Negative if
                sim < unmatched_threshold. Defaults to matched_threshold
                when set to None.
            negatives_lower_than_unmatched: Boolean which defaults to True. If True
                then negative matches are the ones below the unmatched_threshold,
                whereas ignored matches are in between the matched and unmatched
                threshold. If False, then negative matches are in between the matched
                and unmatched threshold, and everything lower than unmatched is ignored.
            force_match_for_each_row: If True, ensures that each row is matched to
                at least one column (which is not guaranteed otherwise if the
                matched_threshold is high). Defaults to False. See
                argmax_matcher_test.testMatcherForceMatch() for an example.

        Raises:
            ValueError: if unmatched_threshold is set but matched_threshold is not set
                or if unmatched_threshold > matched_threshold.
        """
        if (matched_threshold is None) and (unmatched_threshold is not None):
            raise ValueError('Need to also define matched_threshold when unmatched_threshold is defined')
        self._matched_threshold = matched_threshold
        self._unmatched_threshold: float = 0.
        if unmatched_threshold is None:
            self._unmatched_threshold = matched_threshold
        else:
            if unmatched_threshold > matched_threshold:
                raise ValueError('unmatched_threshold needs to be smaller or equal to matched_threshold')
            self._unmatched_threshold = unmatched_threshold
        if not negatives_lower_than_unmatched:
            if self._unmatched_threshold == self._matched_threshold:
                raise ValueError('When negatives are in between matched and unmatched thresholds, these '
                                 'cannot be of equal value. matched: %s, unmatched: %s',
                                 self._matched_threshold, self._unmatched_threshold)
        self._force_match_for_each_row = force_match_for_each_row
        self._negatives_lower_than_unmatched = negatives_lower_than_unmatched

    def _match_when_rows_are_empty(self, similarity_matrix):
        """Performs matching when the rows of similarity matrix are empty.

        When the rows are empty, all detections are false positives. So we return
        a tensor of -1's to indicate that the columns do not match to any rows.

        Returns:
            matches:  int32 tensor indicating the row each column matches to.
        """
        return -1 * torch.ones(similarity_matrix.shape[1], dtype=torch.long, device=similarity_matrix.device)

    def _match_when_rows_are_non_empty(self, similarity_matrix):
        """Performs matching when the rows of similarity matrix are non empty.

        Returns:
            matches:  int32 tensor indicating the row each column matches to.
        """
        # Matches for each column
        matched_vals, matches = torch.max(similarity_matrix, 0)

        # Deal with matched and unmatched threshold
        if self._matched_threshold is not None:
            # Get logical indices of ignored and unmatched columns as tf.int64
            below_unmatched_threshold = self._unmatched_threshold > matched_vals
            between_thresholds = (matched_vals >= self._unmatched_threshold) & \
                                 (self._matched_threshold > matched_vals)

            if self._negatives_lower_than_unmatched:
                matches = self._set_values_using_indicator(matches, below_unmatched_threshold, -1)
                matches = self._set_values_using_indicator(matches, between_thresholds, -2)
            else:
                matches = self._set_values_using_indicator(matches, below_unmatched_threshold, -2)
                matches = self._set_values_using_indicator(matches, between_thresholds, -1)

        if self._force_match_for_each_row:
            force_match_column_ids = torch.argmax(similarity_matrix, 1)
            force_match_column_indicators = one_hot_bool(force_match_column_ids, similarity_matrix.shape[1])
            force_match_column_mask, force_match_row_ids = torch.max(force_match_column_indicators, 0)
            final_matches = torch.where(force_match_column_mask, force_match_row_ids, matches)
            return final_matches
        else:
            return matches

    def match(self, similarity_matrix):
        """Tries to match each column of the similarity matrix to a row.

        Args:
            similarity_matrix: tensor of shape [N, M] representing any similarity metric.

        Returns:
            Match object with corresponding matches for each of M columns.
        """
        if similarity_matrix.shape[0] == 0:
            return Match(self._match_when_rows_are_empty(similarity_matrix))
        else:
            return Match(self._match_when_rows_are_non_empty(similarity_matrix))

    def _set_values_using_indicator(self, x, indicator, val: int):
        """Set the indicated fields of x to val.

        Args:
            x: tensor.
            indicator: boolean with same shape as x.
            val: scalar with value to set.

        Returns:
            modified tensor.
        """
        indicator = indicator.to(dtype=x.dtype)
        return x * (1 - indicator) + val * indicator



# type: ignore

import torch
import torchvision
import numpy as np
import mira.thirdparty.effdet as effdet
import omegaconf
import mira.detectors.common as mdc


class EfficientDet(torch.nn.Module):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        config = omegaconf.OmegaConf.merge(
            effdet.get_efficientdet_config(model_name=MODEL_NAME),
            omegaconf.OmegaConf.create(
                {
                    "min_level": MIN_LEVEL,
                    "max_level": MAX_LEVEL,
                    "num_levels": NUM_LEVELS,
                    "max_detection_points": MAX_DETECTION_POINTS,
                }
            ),
        )
        self.model = effdet.create_model_from_config(
            config=config,
            num_classes=NUM_CLASSES,
            bench_task="",
            pretrained=False,
            pretrained_backbone=False,
            image_size=(INPUT_HEIGHT, INPUT_WIDTH),
            max_det_per_image=MAX_DET_PER_IMAGE,
        )
        self.config = {"anchors": effdet.anchors.Anchors.from_config(self.model.config)}
        self.mean = torch.tensor(
            np.array([[[[0.485]], [[0.456]], [[0.406]]]]), dtype=torch.float32
        )
        self.std = torch.tensor(
            np.array([[[[0.229]], [[0.224]], [[0.225]]]]), dtype=torch.float32
        )

    def normalize(self, x):
        return (x - self.mean.to(x.device)) / self.std.to(x.device)

    def forward(self, x):
        resized, scales, sizes = mdc.resize(
            x,
            resize_method=RESIZE_METHOD,
            height=INPUT_HEIGHT,
            width=INPUT_WIDTH,
            base=128,
        )
        # Go from [sy, sx] to [sx, sy, sx, sy]
        scales = scales[:, [1, 0]].repeat((1, 2))
        sizes = sizes[:, [1, 0]].repeat((1, 2))
        resized = self.normalize(resized)
        if (
            self.model.config.image_size[0] != resized.shape[2]
            or self.model.config.image_size[1] != resized.shape[3]
        ):
            self.model.config = omegaconf.OmegaConf.merge(  # type: ignore
                self.model.config,
                omegaconf.OmegaConf.create(
                    {"image_size": (resized.shape[2], resized.shape[3])}
                ),
            )
            self.config["anchors"] = effdet.anchors.Anchors.from_config(
                self.model.config
            )
        y = self.model(resized)
        class_out, box_out = y
        class_out, box_out, indices, classes = effdet.bench._post_process(
            class_out,
            box_out,
            num_levels=self.model.config.num_levels,
            num_classes=self.model.config.num_classes,
            max_detection_points=self.model.config.max_detection_points,
        )
        detections = effdet.bench._batch_detection(
            class_out.shape[0],
            class_out.cpu(),
            box_out.cpu(),
            self.config["anchors"].boxes.cpu(),
            indices.cpu(),
            classes.cpu(),
            img_scale=None,
            img_size=None,
            max_det_per_image=self.model.config.max_det_per_image,
            soft_nms=True,
        )
        clipped = [
            group[:, :4].min(group_size.unsqueeze(0))
            for group, group_size in zip(detections, sizes)
        ]
        has_area = [((c[:, 3] - c[:, 1]) * (c[:, 2] - c[:, 0])) > 0 for c in clipped]
        return [
            {
                "boxes": boxes[box_has_area] * (1 / scaler),
                "labels": group[box_has_area, 5].type(torch.IntTensor),
                "scores": group[box_has_area, 4],
            }
            for group, boxes, box_has_area, scaler in zip(
                detections, clipped, has_area, scales
            )
        ]



import typing

try:
    import torch
except ImportError:
    torch = None  # type: ignore
import pandas as pd
import typing_extensions as tx

if typing.TYPE_CHECKING:
    from ..detectors.detector import Detector

# pylint: disable=too-few-public-methods
class CallbackProtocol(tx.Protocol):
    """A protocol defining how we expect callbacks to behave."""

    def __call__(
        self,
        detector: "Detector",
        summaries: typing.List[typing.Dict[str, typing.Any]],
    ) -> typing.Dict[str, typing.Any]:
        pass


def best_weights(
    filepath, metric="loss", method="min", key="saved"
) -> CallbackProtocol:
    """A callback that saves the best model weights according to a metric.

    Args:
        metric: The metric to track. Use dot notation for nested attributes
            (e.g., val_mAP.{class_name}).
        method: How to handle the metric ("min" minimizes the metric while "max"
            maximizes it).
        key: What name to use for the saved flag.
    """
    if torch is None:
        raise ValueError("You must install pytorch to use this callback.")

    def callback(detector, summaries):
        saved = False
        summaries_df = pd.json_normalize(summaries)
        best_idx = (
            summaries_df[metric].idxmax()
            if method == "max"
            else summaries_df[metric].idxmin()
        )
        if best_idx == len(summaries_df) - 1:
            torch.save(detector.model.state_dict(), filepath)
            saved = True
        return {key: saved}

    return callback


def csv_logger(filepath) -> CallbackProtocol:
    """A callback that saves a CSV of the summaries to a specific
    filepath.

    Args:
        filepath: The filepath where the logs will be saved.
    """
    # pylint: disable=unused-argument
    def callback(detector, summaries):
        pd.json_normalize(summaries).to_csv(filepath, index=False)
        return {}

    return callback


def mAP(collection, key="val_mAP", **kwargs) -> CallbackProtocol:
    """Build a callback that computes mAP. All kwargs
    passed to detector.mAP()"""

    # pylint: disable=unused-argument
    def callback(detector, summaries):
        return {
            key: {
                k: round(v, 2)
                for k, v in detector.mAP(collection=collection, **kwargs).items()
            }
        }

    return callback



# Copyright 2020 Google Research. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Matcher interface and Match class.

This module defines the Matcher interface and the Match object. The job of the
matcher is to match row and column indices based on the similarity matrix and
other optional parameters. Each column is matched to at most one row. There
are three possibilities for the matching:

1) match: A column matches a row.
2) no_match: A column does not match any row.
3) ignore: A column that is neither 'match' nor no_match.

The ignore case is regularly encountered in object detection: when an anchor has
a relatively small overlap with a ground-truth box, one neither wants to
consider this box a positive example (match) nor a negative example (no match).

The Match class is used to store the match results and it provides simple apis
to query the results.
"""
import torch


@torch.jit.script
class Match(object):
    """Class to store results from the matcher.

    This class is used to store the results from the matcher. It provides
    convenient methods to query the matching results.
    """

    def __init__(self, match_results: torch.Tensor):
        """Constructs a Match object.

        Args:
            match_results: Integer tensor of shape [N] with (1) match_results[i]>=0,
                meaning that column i is matched with row match_results[i].
                (2) match_results[i]=-1, meaning that column i is not matched.
                (3) match_results[i]=-2, meaning that column i is ignored.

        Raises:
            ValueError: if match_results does not have rank 1 or is not an integer int32 scalar tensor
        """
        if len(match_results.shape) != 1:
            raise ValueError('match_results should have rank 1')
        if match_results.dtype not in (torch.int32, torch.int64):
            raise ValueError('match_results should be an int32 or int64 scalar tensor')
        self.match_results = match_results

    def matched_column_indices(self):
        """Returns column indices that match to some row.

        The indices returned by this op are always sorted in increasing order.

        Returns:
            column_indices: int32 tensor of shape [K] with column indices.
        """
        return torch.nonzero(self.match_results > -1).flatten().long()

    def matched_column_indicator(self):
        """Returns column indices that are matched.

        Returns:
            column_indices: int32 tensor of shape [K] with column indices.
        """
        return self.match_results >= 0

    def num_matched_columns(self):
        """Returns number (int32 scalar tensor) of matched columns."""
        return self.matched_column_indices().numel()

    def unmatched_column_indices(self):
        """Returns column indices that do not match any row.

        The indices returned by this op are always sorted in increasing order.

        Returns:
          column_indices: int32 tensor of shape [K] with column indices.
        """
        return torch.nonzero(self.match_results == -1).flatten().long()

    def unmatched_column_indicator(self):
        """Returns column indices that are unmatched.

        Returns:
          column_indices: int32 tensor of shape [K] with column indices.
        """
        return self.match_results == -1

    def num_unmatched_columns(self):
        """Returns number (int32 scalar tensor) of unmatched columns."""
        return self.unmatched_column_indices().numel()

    def ignored_column_indices(self):
        """Returns column indices that are ignored (neither Matched nor Unmatched).

        The indices returned by this op are always sorted in increasing order.

        Returns:
          column_indices: int32 tensor of shape [K] with column indices.
        """
        return torch.nonzero(self.ignored_column_indicator()).flatten().long()

    def ignored_column_indicator(self):
        """Returns boolean column indicator where True means the column is ignored.

        Returns:
            column_indicator: boolean vector which is True for all ignored column indices.
        """
        return self.match_results == -2

    def num_ignored_columns(self):
        """Returns number (int32 scalar tensor) of matched columns."""
        return self.ignored_column_indices().numel()

    def unmatched_or_ignored_column_indices(self):
        """Returns column indices that are unmatched or ignored.

        The indices returned by this op are always sorted in increasing order.

        Returns:
            column_indices: int32 tensor of shape [K] with column indices.
        """
        return torch.nonzero(0 > self.match_results).flatten().long()

    def matched_row_indices(self):
        """Returns row indices that match some column.

        The indices returned by this op are ordered so as to be in correspondence with the output of
        matched_column_indicator().  For example if self.matched_column_indicator() is [0,2],
        and self.matched_row_indices() is [7, 3], then we know that column 0 was matched to row 7 and
        column 2 was matched to row 3.

        Returns:
            row_indices: int32 tensor of shape [K] with row indices.
        """
        return torch.gather(self.match_results, 0, self.matched_column_indices()).flatten().long()

    def gather_based_on_match(self, input_tensor, unmatched_value, ignored_value):
        """Gathers elements from `input_tensor` based on match results.

        For columns that are matched to a row, gathered_tensor[col] is set to input_tensor[match_results[col]].
        For columns that are unmatched, gathered_tensor[col] is set to unmatched_value. Finally, for columns that
        are ignored gathered_tensor[col] is set to ignored_value.

        Note that the input_tensor.shape[1:] must match with unmatched_value.shape
        and ignored_value.shape

        Args:
            input_tensor: Tensor to gather values from.
            unmatched_value: Constant tensor or python scalar value for unmatched columns.
            ignored_value: Constant tensor or python scalar for ignored columns.

        Returns:
            gathered_tensor: A tensor containing values gathered from input_tensor.
                The shape of the gathered tensor is [match_results.shape[0]] + input_tensor.shape[1:].
        """
        if isinstance(ignored_value, torch.Tensor):
            input_tensor = torch.cat([ignored_value, unmatched_value, input_tensor], dim=0)
        else:
            # scalars
            input_tensor = torch.cat([
                torch.tensor([ignored_value, unmatched_value], dtype=input_tensor.dtype, device=input_tensor.device),
                input_tensor], dim=0)
        gather_indices = torch.clamp(self.match_results + 2, min=0)
        gathered_tensor = torch.index_select(input_tensor, 0, gather_indices)
        return gathered_tensor
